import typing as t
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

from .exceptions import MissingDependencyException
from ._internal.types import Tag
from ._internal.runner import Runner
from ._internal.configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from _internal.models import ModelStore

try:
    ...
except ImportError:  # pragma: no cover
    raise MissingDependencyException("")


@inject
def load(
    tag: str,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
):
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `xgboost.core.Booster` from BentoML modelstore.

    Examples::
    """  # noqa


@inject
def save(
    name: str,
    model: t.Any,
    *,
    metadata: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> str:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`xgboost.core.Booster`):
            Instance of model to be saved
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
    """  # noqa


class _PySparkMLlibRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        resource_quota: t.Dict[str, t.Any],
        batch_options: t.Dict[str, t.Any],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(str(tag), resource_quota, batch_options)

    @property
    def required_models(self) -> t.List[Tag]:
        ...

    @property
    def num_concurrency_per_replica(self) -> int:
        ...

    @property
    def num_replica(self) -> int:
        ...

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:
        ...

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _run_batch(self, input_data) -> t.Any:
        ...


@inject
def load_runner(
    tag: t.Union[str, Tag],
    *,
    resource_quota: t.Union[None, t.Dict[str, t.Any]] = None,
    batch_options: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_PySparkMLlibRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.xgboost.load_runner` implements a Runner class that
    wrap around a Xgboost booster model, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.pyspark` model

    Examples::
    """  # noqa
    return _PySparkMLlibRunner(
        tag=tag,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )


# import importlib
# import json
# import logging
# import os
# import typing as t
#
# import bentoml._internal.constants as _const
#
# from ._internal.models.base import MODEL_NAMESPACE, Model
# from ._internal.types import GenericDictType, PathType
# from ._internal.utils import LazyLoader
# from .exceptions import BentoMLException
#
# logger = logging.getLogger(__name__)
#
# _exc = _const.IMPORT_ERROR_MSG.format(
#     fwr="pyspark.mllib",
#     module=__name__,
#     inst="First install Apache Spark, https://spark.apache.org/downloads.html."
#     " Then run `pip install pyspark`",
# )
#
# if t.TYPE_CHECKING:  # pragma: no cover
#     # pylint: disable=unused-import
#     import pyspark
#     import pyspark.ml as ml
#     import pyspark.sql as sql
# else:
#     pyspark = LazyLoader("pyspark", globals(), "pyspark", exc_msg=_exc)
#     ml = LazyLoader("ml", globals(), "pyspark.ml", exc_msg=_exc)
#     sql = LazyLoader("sql", globals(), "pyspark.sql", exc_msg=_exc)
#
# # NOTE: the usage of SPARK_SESSION_NAMESPACE is to provide a consistent session
# #  among imports if users need to use SparkSession.
# SPARK_SESSION_NAMESPACE: str = "PySparkMLlibModel"
#
# DEPRECATION_MLLIB_WARNING: str = """
# {model} is using the older library `pyspark.mllib`.
# Consider to upgrade your model to use `pyspark.ml`.
# BentoML will still try to load {model} with `pyspark.sql.SparkSession`,
# but expect unintended behaviour.
# """
#
#
# class PySparkMLlibModel(Model):
#     """
#     Model class for saving/loading :obj:`pyspark` models
#     using :obj:`pyspark.ml` and :obj:`pyspark.mllib`
#
#     Args:
#         model (`pyspark.ml.Model`):
#             Every PySpark model is of type :obj:`pyspark.ml.Model`
#         spark_session (`pyspark.sql.SparkSession`, `optional`, default to `None`):
#             Optional SparkSession used to load PySpark model representation.
#         metadata (`GenericDictType`,  `optional`, default to `None`):
#             Class metadata
#
#     Raises:
#         MissingDependencyException:
#             :obj:`pyspark` is required by PySparkMLlibModel
#
#     .. WARNING::
#
#         :obj:`spark_session` should only be used when your current model is running
#         older version of `pyspark.ml` (`pyspark.mllib`). Consider to upgrade your mode
#         beforehand to use `pyspark.ml`.
#
#     Example usage under :code:`train.py`::
#
#         TODO:
#
#     One then can define :code:`bento.py`::
#
#         TODO:
#     """
#
#     _model: "pyspark.ml.Model"
#
#     def __init__(
#         self,
#         model: "pyspark.ml.Model",
#         spark_session: t.Optional["pyspark.sql.SparkSession"] = None,
#         metadata: t.Optional[GenericDictType] = None,
#     ):
#         super(PySparkMLlibModel, self).__init__(model, metadata=metadata)
#         # NOTES: referred to docstring, spark_session is mainly used
#         #  for backward compatibility.
#         self._spark_sess = spark_session
#
#     @classmethod
#     def load(cls, path: PathType) -> "pyspark.ml.Model":
#
#         model_path: str = str(os.path.join(path, MODEL_NAMESPACE))
#
#         # NOTE (future ref): A large model metadata might
#         #  comprise of multiple `part` files, instead of assigning,
#         #  loop through the directory.
#         metadata_path: str = str(os.path.join(model_path, "metadata/part-00000"))
#
#         try:
#             with open(metadata_path, "r") as meta_file:
#                 metadata = json.load(meta_file)
#         except IOError:
#             raise BentoMLException(
#                 "Incorrectly serialized model was loaded. Unable to load metadata"
#             )
#         if "class" not in metadata:
#             raise BentoMLException("malformed metadata file.")
#         model_class = metadata["class"]
#
#         # process imports from metadata
#         stripped_apache_module: t.List[str] = model_class.split(".")[2:]
#         py_module = "py" + ".".join(stripped_apache_module[:-1])  # skip org.apache
#         class_name = stripped_apache_module[-1]
#
#         loaded_model = getattr(importlib.import_module(py_module), class_name)
#         if not issubclass(loaded_model, ml.Model):
#             logger.warning(DEPRECATION_MLLIB_WARNING.format(model=loaded_model))
#             _spark_sess = sql.SparkSession.builder.appName(
#                 SPARK_SESSION_NAMESPACE
#             ).getOrCreate()
#             model = loaded_model.load(_spark_sess.sparkContext, model_path)
#         else:
#             model = loaded_model.load(model_path)
#
#         return model
#
#     def save(self, path: PathType) -> None:
#         if not isinstance(self._model, ml.Model):
#             logger.warning(DEPRECATION_MLLIB_WARNING.format(model=self._model))
#             self._model.save(self._spark_sess, os.path.join(path, MODEL_NAMESPACE))
#         else:
#             self._model.save(os.path.join(path, MODEL_NAMESPACE))



import pickle
import typing as t
import zipfile
import functools
from typing import TYPE_CHECKING
from pathlib import Path

import cloudpickle
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner

from ..models import Model
from ..models import PT_EXT
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ...exceptions import MissingDependencyException
from ..runner.utils import Params
from ..runner.utils import TypeRef
from ..runner.container import Payload
from ..runner.container import DataContainer
from ..runner.container import DataContainerRegistry
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ..models import ModelStore

try:
    import numpy as np
    import torch
    import torch.nn.parallel as parallel
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """
        torch is required in order to use module `bentoml.pytorch`.
         Instruction: Refers to https://pytorch.org/get-started/locally/
         to setup PyTorch correctly.
        """  # noqa
    )

_ModelType = t.Union["torch.nn.Module", "torch.jit.ScriptModule"]  # type: ignore[reportPrivateUsage]

_torch_version = get_pkg_version("torch")


def _is_gpu_available() -> bool:  # pragma: no cover
    return torch.cuda.is_available()


@inject
def load(
    tag: t.Union[str, Tag],
    device_id: t.Optional[str] = "cpu",
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> _ModelType:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        device_id (`str`, `optional`):
            Optional devices to put the given model on. Refers to https://pytorch.org/docs/stable/tensor_attributes.html#torch.torch.device
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of either `torch.jit.ScriptModule` or `torch.nn.Module` from BentoML modelstore.

    Examples::
        import bentoml.pytorch
        booster = bentoml.pytorch.load(
            'lit_classifier:20201012_DE43A2', device_id="cuda:0")
    """  # noqa
    bentoml_model = model_store.get(tag)
    weight_file = bentoml_model.path_of(f"{SAVE_NAMESPACE}{PT_EXT}")
    # TorchScript Models are saved as zip files
    if zipfile.is_zipfile(weight_file):
        model: "torch.jit.ScriptModule" = torch.jit.load(weight_file, map_location=device_id)  # type: ignore[reportPrivateImportUsage] # noqa: LN001
        return model
    else:
        with Path(weight_file).open("rb") as file:
            return cloudpickle.load(file)


@inject
def save(
    name: str,
    model: _ModelType,
    *,
    metadata: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`t.Union[torch.nn.Module, torch.jit.ScriptModule]`):
            Instance of model to be saved
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        import bentoml.pytorch
        import torch

        class NGramLanguageModeler(nn.Module):

            def __init__(self, vocab_size, embedding_dim, context_size):
                super(NGramLanguageModeler, self).__init__()
                self.embeddings = nn.Embedding(vocab_size, embedding_dim)
                self.linear1 = nn.Linear(context_size * embedding_dim, 128)
                self.linear2 = nn.Linear(128, vocab_size)

            def forward(self, inputs):
                embeds = self.embeddings(inputs).view((1, -1))
                out = F.relu(self.linear1(embeds))
                out = self.linear2(out)
                log_probs = F.log_softmax(out, dim=1)
                return log_probs

        tag = bentoml.pytorch.save("ngrams", NGramLanguageModeler(len(vocab), EMBEDDING_DIM, CONTEXT_SIZE))
        # example tag: ngrams:20201012_DE43A2

    Integration with Torch Hub and BentoML::
        import bentoml.pytorch
        import torch

        resnet50 = torch.hub.load("pytorch/vision", "resnet50", pretrained=True)
        ...
        # trained a custom resnet50

        tag = bentoml.pytorch.save("resnet50", resnet50)
    """  # noqa
    context: t.Dict[str, t.Any] = {
        "framework_name": "torch",
        "pip_dependencies": [f"torch=={_torch_version}"],
    }
    _model = Model.create(
        name,
        module=__name__,
        options=None,
        context=context,
        metadata=metadata,
    )
    weight_file = _model.path_of(f"{SAVE_NAMESPACE}{PT_EXT}")
    if isinstance(model, torch.jit.ScriptModule):  # type: ignore[reportPrivateUsage]
        torch.jit.save(model, weight_file)  # type: ignore[reportUnknownMemberType]
    else:
        with open(weight_file, "wb") as file:
            cloudpickle.dump(model, file)

    _model.save(model_store)
    return _model.tag


class _PyTorchRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        device_id: str,
        partial_kwargs: t.Optional[t.Dict[str, t.Any]],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        in_store_tag = model_store.get(tag).tag

        super().__init__(str(in_store_tag), resource_quota, batch_options)
        self._predict_fn_name = predict_fn_name
        self._model_store = model_store
        if "cuda" in device_id:
            try:
                _, dev = device_id.split(":")
                self.resource_quota.gpus = [dev]
            except ValueError:
                self.resource_quota.gpus = [
                    str(i) for i in range(torch.cuda.device_count())
                ]
        self._device_id = device_id
        self._partial_kwargs = partial_kwargs or dict()
        self._tag = in_store_tag

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        if _is_gpu_available() and self.resource_quota.on_gpu:
            return 1
        return int(round(self.resource_quota.cpu))

    @property
    def num_replica(self) -> int:
        if _is_gpu_available() and self.resource_quota.on_gpu:
            return torch.cuda.device_count()
        return 1

    def _configure(self) -> None:
        torch.set_num_threads(self.num_concurrency_per_replica)
        if self.resource_quota.on_gpu and _is_gpu_available():
            torch.set_default_tensor_type("torch.cuda.FloatTensor")
        else:
            torch.set_default_tensor_type("torch.FloatTensor")

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    @torch.no_grad()
    def _setup(self) -> None:  # type: ignore[override]
        self._configure()
        if self.resource_quota.on_gpu and _is_gpu_available():
            self._model = parallel.DataParallel(
                load(
                    self._tag,
                    model_store=self._model_store,
                    device_id=self._device_id,
                ),
            )
            torch.cuda.empty_cache()
        else:
            self._model = load(
                self._tag,
                model_store=self._model_store,
                device_id=self._device_id,
            )
        raw_predict_fn = getattr(self._model, self._predict_fn_name)
        self._predict_fn: t.Callable[..., torch.Tensor] = functools.partial(
            raw_predict_fn, **self._partial_kwargs
        )

    @torch.no_grad()
    def _run_batch(
        self,
        *args: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor],
        **kwargs: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor],
    ) -> torch.Tensor:

        params = Params[t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor]](
            *args, **kwargs
        )

        def _mapping(
            item: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor]
        ) -> torch.Tensor:
            if isinstance(item, np.ndarray):
                item = torch.from_numpy(item)
            if self.resource_quota.on_gpu:
                item = item.cuda()
            return item

        params = params.map(_mapping)

        # inference mode is required for PyTorch version 1.9.*
        if _torch_version.startswith("1.9"):
            with torch.inference_mode():
                res = self._predict_fn(*params.args, **kwargs)
        else:
            res = self._predict_fn(*params.args, **kwargs)
        return res


@inject
def load_runner(
    tag: t.Union[str, Tag],
    *,
    predict_fn_name: str = "__call__",
    device_id: str = "cpu:0",
    partial_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
    batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_PyTorchRunner":
    """
        Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.pytorch.load_runner` implements a Runner class that
    wrap around a pytorch instance, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `__call__`):
            inference function to be used.
        device_id (`t.Union[str, int, t.List[t.Union[str, int]]]`, `optional`, default to `cpu`):
            Optional devices to put the given model on. Refers to https://pytorch.org/docs/stable/tensor_attributes.html#torch.torch.device
        partial_kwargs (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Common kwargs passed to model for this runner
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.pytorch` model

    Examples::
        import bentoml.pytorch
        runner = bentoml.pytorch.load_runner("ngrams:20201012_DE43A2")
        runner.run(pd.DataFrame("/path/to/csv"))
    """  # noqa
    return _PyTorchRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        device_id=device_id,
        partial_kwargs=partial_kwargs,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )


class PytorchTensorContainer(DataContainer[torch.Tensor, torch.Tensor]):
    @classmethod
    def singles_to_batch(cls, singles, batch_axis=0):
        return torch.stack(singles, dim=batch_axis)

    @classmethod
    def batch_to_singles(cls, batch, batch_axis=0):
        return [
            torch.squeeze(tensor, dim=batch_axis)
            for tensor in torch.split(batch, 1, dim=batch_axis)
        ]

    @classmethod
    @inject
    def single_to_payload(
        cls,
        single_data,
    ) -> Payload:
        return cls.create_payload(
            pickle.dumps(single_data),
            {"plasma": False},
        )

    @classmethod
    @inject
    def payload_to_single(
        cls,
        payload: Payload,
    ):
        return pickle.loads(payload.data)

    batch_to_payload = single_to_payload
    payload_to_batch = payload_to_single


DataContainerRegistry.register_container(
    TypeRef("torch", "Tensor"),
    TypeRef("torch", "Tensor"),
    PytorchTensorContainer,
)



import re
import sys
import typing as t
import logging

from ...exceptions import BentoMLException
from ..configuration import BENTOML_VERSION
from ..configuration import is_pypi_installed_bentoml

logger = logging.getLogger(__name__)

SUPPORTED_PYTHON_VERSION: t.List[str] = ["3.7", "3.8", "3.9"]
SUPPORTED_BASE_DISTROS: t.List[str] = ["slim", "centos7", "centos8"]
SUPPORTED_GPU_DISTROS: t.List[str] = SUPPORTED_BASE_DISTROS

SUPPORTED_RELEASES_COMBINATION: t.Dict[str, t.List[str]] = {
    "cudnn": SUPPORTED_BASE_DISTROS,
    "devel": SUPPORTED_BASE_DISTROS,
    "runtime": SUPPORTED_BASE_DISTROS + ["ami2", "alpine3.14"],
}

SEMVER_REGEX = re.compile(r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)?$")

BACKWARD_COMPATIBILITY_WARNING: str = """\
Since 1.0.0, we changed the format of docker tags, thus {classname}
will only supports image tag from 1.0.0 forward, while detected bentoml_version
is {bentoml_version}.
Refers to https://hub.docker.com/r/bentoml/model-server/
if you need older version of bentoml. Using default devel image instead...
"""  # noqa: E501

RELEASE_FORMAT = (
    "bentoml/bento-server:{release_type}-python{python_version}-{distro}{suffix}"
)


class ImageProvider(object):
    """
    BentoML-provided docker images.

    Args:
        distro (string): distro that will be used for docker image.
        python_version (string): python version for given docker image.
        gpu (bool): Optional GPU supports.

    Raises:
         BentoMLException: ``python_version`` and ``distro`` \
                        is not yet supported by BentoML
         BentoMLException: invalid combination of supported GPU distros

    Returns:

        .. code-block:: shell

            bentoml/bento-server:<release_type>-<python_version>-<distro>-<suffix>

            # suffix: runtime or cudnn
            # distro: formatted <distro><version> (ami2, centos8, centos7)
            # python_version: 3.7,3.8
            # release_type: devel or versioned (0.14.0)

        Example results:

        * bentoml/bento-server:`devel-python3.7-slim`
        * bentoml/bento-server:`0.14.0-python3.8-centos8-cudnn`
        * bentoml/bento-server:`0.14.0-python3.7-ami2-runtime`


    Example::

    from bento import svc

    svc.build(
        ...
        env=dict(
            docker=dict(
                distro='debian',
                python='3.8',
                gpu=True
            )
        )
    )
    """

    def __init__(
        self,
        distro: t.Optional[
            't.Literal["slim", "amazonlinux2", "alpine", "centos7", "centos8"]'
        ] = None,
        python_version: t.Optional[str] = None,
        gpu: t.Optional[bool] = None,
    ) -> None:
        if distro is None:
            distro = "slim"
        if gpu is None:
            gpu = False

        if is_pypi_installed_bentoml():
            self._release_type = BENTOML_VERSION
            self._suffix = "cudnn" if gpu else "runtime"
        else:
            self._release_type: str = "devel"
            self._suffix = ""

        if python_version:
            assert re.match(r"^[2,3]\.[0-9]{1,2}$", python_version), (
                f"python_version '{python_version}' should be in format "
                "'{major}.{minor}', e.g. 'python_version=\"3.7\"'"
            )
        else:
            logger.debug(f"No python_version is specified, using {python_version}.")
            python_version = ".".join(map(str, sys.version_info[:2]))

        if python_version not in SUPPORTED_PYTHON_VERSION:
            raise BentoMLException(
                f"BentoML does not provide docker image for Python {python_version}."
                f"Supported Python versions are {SUPPORTED_PYTHON_VERSION}. "
                f"You may specify a different python version for BentoML to use"
                "e.g: \"env=dict(distro='centos7', python_version='3.7')\""
            )

        self._python_version: str = python_version

        if gpu and distro not in SUPPORTED_GPU_DISTROS:
            raise BentoMLException(
                f"distro '{distro}' with GPU={gpu} is not supported. "
                f"GPU-supported distros are: {SUPPORTED_GPU_DISTROS} "
            )

        if (
            distro not in SUPPORTED_RELEASES_COMBINATION["devel"]
            and self._release_type == "devel"
        ):
            raise BentoMLException(f"{distro} doesn't support devel tags.")

        if self._suffix and distro not in SUPPORTED_RELEASES_COMBINATION[self._suffix]:
            raise BentoMLException(
                f"{distro} is not yet supported. "
                "Supported distros: "
                f"{SUPPORTED_RELEASES_COMBINATION[self._suffix]}"
            )

        self._distro: str = distro

    def __repr__(self):
        actual_suffix = "-" + self._suffix if self._suffix else ""
        return RELEASE_FORMAT.format(
            release_type=self._release_type,
            python_version=self._python_version,
            distro=self._distro,
            suffix=actual_suffix,
        )



import re

from ...exceptions import InvalidArgument

tag_fmt = "[a-z0-9]([-._a-z0-9]*[a-z0-9])?"
tag_max_length = 63
tag_max_length_error_msg = (
    "a tag's name or version must be at most {tag_max_length} characters in length"
)
tag_invalid_error_msg = "a tag's name or version must consist of alphanumeric characters, '_', '-', or '.', and must start and end with an alphanumeric character"
tag_regex = re.compile(f"^{tag_fmt}$")


def validate_tag_str(value: str):
    """
    Validate that a tag value (either name or version) is a simple string that:
        * Must be at most 63 characters long,
        * Begin and end with an alphanumeric character (`[a-z0-9A-Z]`), and
        * May contain dashes (`-`), underscores (`_`) dots (`.`), or alphanumerics
          between.
    """
    errors = []
    if len(value) > tag_max_length:
        errors.append(tag_max_length_error_msg)
    if tag_regex.match(value) is None:
        errors.append(tag_invalid_error_msg)

    if errors:
        raise InvalidArgument(f"{value} is not a valid tag: " + ", and ".join(errors))



import typing as t
from typing import TYPE_CHECKING

import numpy as np
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import Model
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    import pandas as pd

    from ..models import ModelStore  # noqa


try:
    import h2o
    import h2o.model
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """h2o is required in order to use module `bentoml.h2o`, install h2o
        with `pip install h2o`. For more information, refers to
        https://docs.h2o.ai/h2o/latest-stable/h2o-docs/downloading.html#install-in-python
        """
    )

_h2o_version = get_pkg_version("h2o")


@inject
def load(
    tag: t.Union[str, Tag],
    init_params: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> h2o.model.model_base.ModelBase:
    """
    Load a model from BentoML local modelstore with given tag.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        init_params (`t.Dict[str, t.Union[str, t.Any]]`):
            Params for h2o server initialization
         model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `h2o.model.model_base.ModelBase`

    Examples::
        TODO
    """  # noqa

    if not init_params:
        init_params = dict()

    h2o.init(**init_params)

    model = model_store.get(tag)
    if model.info.module != __name__:
        raise BentoMLException(  # pragma: no cover
            f"Model {tag} was saved with"
            f" module {model.info.module},"
            f" failed loading with {__name__}."
        )

    path = model.path_of(SAVE_NAMESPACE)
    h2o.no_progress()
    return h2o.load_model(path)


@inject
def save(
    name: str,
    model: h2o.model.model_base.ModelBase,
    *,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`h2o.model.model_base.ModelBase`):
            Instance of h2o model to be saved.
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples:
        TODO


    """  # noqa

    context: t.Dict[str, t.Any] = {
        "framework_name": "h2o",
        "pip_dependencies": [f"h2o=={_h2o_version}"],
    }
    options: t.Dict[str, t.Any] = dict()

    _model = Model.create(
        name,
        module=__name__,
        options=options,
        context=context,
        metadata=metadata,
    )

    h2o.save_model(model=model, path=_model.path, force=True, filename=SAVE_NAMESPACE)

    _model.save(model_store)

    return _model.tag


class _H2ORunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        init_params: t.Optional[t.Dict[str, t.Union[str, t.Any]]],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(str(tag), resource_quota, batch_options)

        self._tag = Tag.from_taglike(tag)
        self._predict_fn_name = predict_fn_name
        self._init_params = init_params
        self._model_store = model_store

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        nthreads = int(self._init_params.get("nthreads", -1))

        if nthreads == -1:
            return int(round(self.resource_quota.cpu))
        return nthreads

    @property
    def num_replica(self) -> int:
        return 1

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:  # type: ignore[override]
        self._model = load(
            self._tag, init_params=self._init_params, model_store=self._model_store
        )
        self._predict_fn = getattr(self._model, self._predict_fn_name)

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[override]
        self, input_data: t.Union[np.ndarray, "pd.DataFrame", h2o.H2OFrame]
    ) -> np.ndarray:
        if not isinstance(input_data, h2o.H2OFrame):
            input_data = h2o.H2OFrame(input_data)
        res = self._predict_fn(input_data)

        if isinstance(res, h2o.H2OFrame):
            res = res.as_data_frame()
        return np.asarray(res)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    predict_fn_name: str = "predict",
    *,
    init_params: t.Optional[t.Dict[str, t.Union[str, t.Any]]],
    resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
    batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> _H2ORunner:
    """Runner represents a unit of serving logic that can be scaled
    horizontally to maximize throughput. `bentoml.h2o.load_runner`
    implements a Runner class that wrap around a h2o model, which
    optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `predict`):
            Options for inference functions. Default to `predict`
        init_params (`t.Dict[str, t.Union[str, t.Any]]`, default to `None`):
            Parameters for h2o.init(). Refers to https://docs.h2o.ai/h2o/latest-stable/h2o-docs/starting-h2o.html#from-python
             for more information
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.h2o`

    Examples::
        TODO

    """  # noqa
    return _H2ORunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        init_params=init_params,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



import typing as t
from typing import TYPE_CHECKING
from contextlib import contextmanager

import fs
from simple_di import inject
from simple_di import Provide

from ._internal.types import Tag
from ._internal.models import Model
from ._internal.configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ._internal.models import ModelStore
    from ._internal.runner import Runner


@inject
def list(  # pylint: disable=redefined-builtin
    tag: t.Optional[t.Union[Tag, str]] = None,
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> t.List["Model"]:
    return _model_store.list(tag)


@inject
def get(
    tag: t.Union[Tag, str],
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "Model":
    return _model_store.get(tag)


@inject
def delete(
    tag: t.Union[Tag, str],
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
):
    _model_store.delete(tag)


@inject
def import_model(
    path: str, *, _model_store: "ModelStore" = Provide[BentoMLContainer.model_store]
) -> Model:
    return Model.from_fs(fs.open_fs(path)).save(_model_store)


def export_model(
    tag: t.Union[Tag, str],
    path: str,
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
):
    model = get(tag, _model_store=_model_store)
    model.export(path)


def push(
    tag: t.Union[Tag, str],
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
):
    raise NotImplementedError


def pull(
    tag: t.Union[Tag, str],
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Model:
    raise NotImplementedError


@inject
@contextmanager
def create(
    name: str,
    *,
    module: str = "",
    labels: t.Optional[t.Dict[str, t.Any]] = None,
    options: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    context: t.Optional[t.Dict[str, t.Any]] = None,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> t.Iterator[Model]:
    res = Model.create(
        name,
        module=module,
        labels=labels,
        options=options,
        metadata=metadata,
        context=context,
    )
    try:
        yield res
    finally:
        res.save(_model_store)


@inject
def load_runner(
    tag: t.Union[Tag, str],
    *,
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "Runner":
    model = get(tag, _model_store=_model_store)
    return model.load_runner()


__all__ = [
    "list",
    "get",
    "delete",
    "import_model",
    "export_model",
    "push",
    "pull",
    "load_runner",
]



# flake8: noqa: E402
from typing import TYPE_CHECKING

from ._internal.configuration import BENTOML_VERSION as __version__
from ._internal.configuration import load_global_config

# Inject dependencies and configurations
load_global_config()

from ._internal.log import configure_logging  # noqa: E402

configure_logging()

from . import models

# bento APIs are top-level
from .bentos import get
from .bentos import list  # pylint: disable=W0622
from .bentos import build
from .bentos import delete
from .bentos import export_bento
from .bentos import import_bento
from ._internal.types import Tag
from ._internal.utils import LazyLoader as _LazyLoader
from ._internal.runner import Runner
from ._internal.runner import SimpleRunner
from ._internal.service import Service
from ._internal.service.loader import load

if TYPE_CHECKING:
    from bentoml import h2o
    from bentoml import flax
    from bentoml import onnx
    from bentoml import gluon
    from bentoml import keras
    from bentoml import spacy
    from bentoml import mlflow
    from bentoml import paddle
    from bentoml import easyocr
    from bentoml import pycaret
    from bentoml import pytorch
    from bentoml import sklearn
    from bentoml import xgboost
    from bentoml import catboost
    from bentoml import lightgbm
    from bentoml import onnxmlir
    from bentoml import detectron
    from bentoml import tensorflow
    from bentoml import statsmodels
    from bentoml import transformers
    from bentoml import pytorch_lightning
else:
    catboost = _LazyLoader("bentoml.catboost", globals(), "bentoml.catboost")
    detectron = _LazyLoader("bentoml.detectron", globals(), "bentoml.detectron")
    easyocr = _LazyLoader("bentoml.easyocr", globals(), "bentoml.easyocr")
    flax = _LazyLoader("bentoml.flax", globals(), "bentoml.flax")
    gluon = _LazyLoader("bentoml.gluon", globals(), "bentoml.gluon")
    h2o = _LazyLoader("bentoml.h2o", globals(), "bentoml.h2o")
    lightgbm = _LazyLoader("bentoml.lightgbm", globals(), "bentoml.lightgbm")
    mlflow = _LazyLoader("bentoml.mlflow", globals(), "bentoml.mlflow")
    onnx = _LazyLoader("bentoml.onnx", globals(), "bentoml.onnx")
    onnxmlir = _LazyLoader("bentoml.onnxmlir", globals(), "bentoml.onnxmlir")
    keras = _LazyLoader("bentoml.keras", globals(), "bentoml.keras")
    paddle = _LazyLoader("bentoml.paddle", globals(), "bentoml.paddle")
    pycaret = _LazyLoader("bentoml.pycaret", globals(), "bentoml.pycaret")
    pytorch = _LazyLoader("bentoml.pytorch", globals(), "bentoml.pytorch")
    pytorch_lightning = _LazyLoader(
        "bentoml.pytorch_lightning", globals(), "bentoml.pytorch_lightning"
    )
    sklearn = _LazyLoader("bentoml.sklearn", globals(), "bentoml.sklearn")
    spacy = _LazyLoader("bentoml.spacy", globals(), "bentoml.spacy")
    statsmodels = _LazyLoader("bentoml.statsmodels", globals(), "bentoml.statsmodels")
    tensorflow = _LazyLoader("bentoml.tensorflow", globals(), "bentoml.tensorflow")
    transformers = _LazyLoader(
        "bentoml.transformers", globals(), "bentoml.transformers"
    )
    xgboost = _LazyLoader("bentoml.xgboost", globals(), "bentoml.xgboost")

__all__ = [
    "__version__",
    "Service",
    "models",
    "Tag",
    "Runner",
    "SimpleRunner",
    # bento APIs
    "list",
    "get",
    "delete",
    "import_bento",
    "export_bento",
    "build",
    "load",
    # frameworks
    "catboost",
    "detectron",
    "easyocr",
    "flax",
    "gluon",
    "h2o",
    "lightgbm",
    "mlflow",
    "onnx",
    "onnxmlir",
    "paddle",
    "pycaret",
    "pytorch",
    "pytorch_lightning",
    "keras",
    "sklearn",
    "spacy",
    "statsmodels",
    "tensorflow",
    "transformers",
    "xgboost",
]



import os
import re
import enum
import typing as t
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

import attr
import psutil

from .utils import cpu_converter
from .utils import gpu_converter
from .utils import mem_converter
from .utils import query_cgroup_cpu_count
from ..types import Tag
from ..configuration.containers import BentoServerContainer

if TYPE_CHECKING:
    import platform

    if platform.system() == "Darwin":
        from psutil._psosx import svmem
    elif platform.system() == "Linux":
        from psutil._pslinux import svmem
    else:
        from psutil._pswindows import svmem


@attr.define
class ResourceQuota:
    cpu: float = attr.field(converter=cpu_converter)
    mem: int = attr.field(converter=mem_converter)

    # Example gpus value: "all", 2, "device=1,2"
    # Default to "None", returns all available GPU devices in current environment
    gpus: t.List[str] = attr.field(converter=gpu_converter, default=None)

    @cpu.default  # type: ignore
    def _get_default_cpu(self) -> float:
        # Default to the total CPU count available in current node or cgroup
        if psutil.POSIX:
            return query_cgroup_cpu_count()
        else:
            cpu_count = os.cpu_count()
            if cpu_count is not None:
                return float(cpu_count)
            raise ValueError("CPU count is NoneType")

    @mem.default  # type: ignore
    def _get_default_mem(self) -> int:
        # Default to the total memory available
        from psutil import virtual_memory

        mem: "svmem" = virtual_memory()
        return mem.total

    @property
    def on_gpu(self) -> bool:
        if self.gpus is not None:
            return len(self.gpus) > 0
        return False


@attr.s
class BatchOptions:
    enabled = attr.ib(
        type=bool,
        default=attr.Factory(BentoServerContainer.config.batch_options.enabled.get),
    )
    max_batch_size = attr.ib(
        type=int,
        default=attr.Factory(
            BentoServerContainer.config.batch_options.max_batch_size.get
        ),
    )
    max_latency_ms = attr.ib(
        type=int,
        default=attr.Factory(
            BentoServerContainer.config.batch_options.max_latency_ms.get
        ),
    )
    input_batch_axis = attr.ib(type=int, default=0)
    output_batch_axis = attr.ib(type=int, default=0)


VARNAME_RE = re.compile(r"\W|^(?=\d)")


class _BaseRunner:
    EXIST_NAMES: t.Set[str] = set()

    def __init__(
        self,
        display_name: t.Union[str, Tag],
        resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
        batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    ):
        # probe an unique name
        if isinstance(display_name, Tag):
            display_name = display_name.name
        if not display_name.isidentifier():
            display_name = VARNAME_RE.sub("_", display_name)
        i = 0
        while True:
            name = display_name if i == 0 else f"{display_name}_{i}"
            if name not in self.EXIST_NAMES:
                self.EXIST_NAMES.add(name)
                break
            else:
                i += 1
        self.name = name

        self.resource_quota = ResourceQuota(
            **(resource_quota if resource_quota else {})
        )
        self.batch_options = BatchOptions(**(batch_options if batch_options else {}))

    @property
    def num_concurrency_per_replica(self) -> int:
        return 1

    @property
    def num_replica(self) -> int:
        return 1

    @property
    def required_models(self) -> t.List[Tag]:
        return []

    @abstractmethod
    def _setup(self) -> None:
        ...

    @property
    def _impl(self) -> "RunnerImpl":
        return RunnerImplPool.get_by_runner(self)

    async def async_run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return await self._impl.async_run(*args, **kwargs)

    async def async_run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return await self._impl.async_run_batch(*args, **kwargs)

    def run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self._impl.run(*args, **kwargs)

    def run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self._impl.run_batch(*args, **kwargs)


class Runner(_BaseRunner, ABC):
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. This Runner class is an abstract class, used for creating
    actual Runners, by implementing `__init__`, `_setup` and `_run_batch` method.

    Runner instances exposes `run` and `run_batch` method, which will eventually be
    piped to the `_run_batch` implementation of the Runner. BentoML applies dynamic
    batching optimization for all Runners by default.

    Why use Runner:
    - Runner allow BentoML to better leverage multiple threads or processes for higher
        hardware utilization (CPU, GPU, Memory)
    - Runner enables higher concurrency in serving workload, which minimizes latency:
        you may prefetch data while model is being executed, parallelize data
        extraction, data transformation or multiple runner execution.
    - Runner comes with dynamic batching optimization, which groups `run` calls into
        batch execution when serving online, the batch size and wait time is adaptive
        to the workload. This can bring massive throughput improvement to a ML service.

    All `_run_batch` argument value must be one of the three types below:
        numpy.ndarray, pandas.DataFrame, List[PickleSerializable]

    Return value of `_run_batch` acceptable types :
        numpy.ndarray, pandas.DataFrame, pandas.Series, List[PickleSerializable]
        Or Tuple of the types above, indicating multiple return values

    Runner `run` accepts argument value of the following types:
        numpy.ndarray => numpy.ndarray
        pandas.DataFrame, pandas.Series => pandas.DataFrame
        any => List[PickleSerializable]

    Note: for pandas.DataFrame and List, the batch_axis must be 0
    """

    @abstractmethod
    def _run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        ...


class SimpleRunner(_BaseRunner, ABC):
    """
    SimpleRunner is a special type of Runner that does not support dynamic batching.
    Instead of `_run_batch` in Runner, a `_run` method is expected to be defined in its
    subclasses.

    A SimpleRunner only exposes `run` method to its users.
        `SimpleRunner._run` can accept arbitrary input type that are pickle-serializable
    """

    @abstractmethod
    def _run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        ...


class RunnerState(enum.IntEnum):
    INIT = 0
    SETTING = 1
    SET = 2


class RunnerImpl:
    def __init__(self, runner: _BaseRunner):
        self._runner = runner
        self._state: RunnerState = RunnerState.INIT

    def setup(self) -> None:
        pass

    @abstractmethod
    async def async_run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        ...

    @abstractmethod
    async def async_run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        ...

    @abstractmethod
    def run(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        ...

    @abstractmethod
    def run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        ...


class RunnerImplPool:
    _runner_map: t.Dict[str, RunnerImpl] = {}

    @classmethod
    def get_by_runner(cls, runner: _BaseRunner) -> RunnerImpl:
        if runner.name in cls._runner_map:
            return cls._runner_map[runner.name]

        remote_runner_mapping = BentoServerContainer.remote_runner_mapping.get()
        if runner.name in remote_runner_mapping:
            from .remote import RemoteRunnerClient

            impl = RemoteRunnerClient(runner)
        else:
            from .local import LocalRunner

            impl = LocalRunner(runner)

        cls._runner_map[runner.name] = impl
        return impl



from ._internal.frameworks.mlflow import load
from ._internal.frameworks.mlflow import save
from ._internal.frameworks.mlflow import load_runner
from ._internal.frameworks.mlflow import import_from_uri

__all__ = ["import_from_uri", "load", "load_runner", "save"]



import typing as t
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..types import PathType
from ..utils import LazyLoader
from ..models import Model
from ..models import PKL_EXT
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    import numpy as np
    from sklearn.base import BaseEstimator
    from sklearn.pipeline import Pipeline
    from pandas.core.frame import DataFrame

    from ..models import ModelStore

try:
    import joblib
    from joblib import parallel_backend
except ImportError:  # pragma: no cover
    try:
        from sklearn.utils._joblib import joblib
        from sklearn.utils._joblib import parallel_backend
    except ImportError:
        raise MissingDependencyException(
            """sklearn is required in order to use the module `bentoml.sklearn`, install
             sklearn with `pip install sklearn`. For more information, refer to
             https://scikit-learn.org/stable/install.html
            """
        )

_sklearn_version = get_pkg_version("scikit-learn")

np = LazyLoader("np", globals(), "numpy")  # noqa: F811
pd = LazyLoader("pd", globals(), "pandas")


def _get_model_info(
    tag: t.Union[str, Tag], model_store: "ModelStore"
) -> t.Tuple["Model", PathType]:
    model = model_store.get(tag)
    if model.info.module != __name__:
        raise BentoMLException(  # pragma: no cover
            f"Model {tag} was saved with module {model.info.module}, failed loading"
            f" with {__name__}."
        )
    model_file = model.path_of(f"{SAVE_NAMESPACE}{PKL_EXT}")

    return model, model_file


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> t.Union["BaseEstimator", "Pipeline"]:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of sklearn model from BentoML modelstore.

    Examples:
        import bentoml.sklearn
        sklearn = bentoml.sklearn.load('my_model:20201012_DE43A2')

    """  # noqa
    _, model_file = _get_model_info(tag, model_store)
    return joblib.load(model_file)


@inject
def save(
    name: str,
    model: t.Union["BaseEstimator", "Pipeline"],
    *,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (``):
            Instance of model to be saved
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples:

    """  # noqa
    context = {
        "framework_name": "sklearn",
        "pip_dependencies": [f"scikit-learn=={_sklearn_version}"],
    }

    _model = Model.create(
        name,
        module=__name__,
        metadata=metadata,
        context=context,
    )

    joblib.dump(model, _model.path_of(f"{SAVE_NAMESPACE}{PKL_EXT}"))

    _model.save(model_store)
    return _model.tag


class _SklearnRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        function_name: str,
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(f"{tag}-{function_name}", resource_quota, batch_options)
        model_info, model_file = _get_model_info(tag, model_store)
        self._model_store = model_store
        self._model_info = model_info
        self._model_file = model_file
        self._backend = "loky"
        self._function_name = function_name

    @property
    def num_concurrency_per_replica(self) -> int:
        return int(round(self.resource_quota.cpu))

    @property
    def num_replica(self) -> int:
        return 1

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._model_info.tag]

    # pylint: disable=attribute-defined-outside-init
    def _setup(self) -> None:
        self._model = joblib.load(filename=self._model_file)
        self._infer_func = getattr(self._model, self._function_name)

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        inputs: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", "DataFrame"],
    ) -> "np.ndarray[t.Any, np.dtype[t.Any]]":
        with parallel_backend(self._backend, n_jobs=self.num_concurrency_per_replica):
            return self._infer_func(inputs)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    function_name: str = "predict",
    *,
    resource_quota: t.Union[None, t.Dict[str, t.Any]] = None,
    batch_options: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_SklearnRunner":

    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.sklearn.load_runner` implements a Runner class that
    wrap around a Sklearn joblib model, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore.
        function_name (`str`, `optional`, default to `predict`):
            Predict function used by a given sklearn model.
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for the target `bentoml.sklearn` model

    Examples::
        import bentoml
        import bentoml.sklearn
        import numpy as np

        from bentoml.io import NumpyNdarray

        input_data = NumpyNdarray()
        runner = bentoml.sklearn.load_runner("my_model:20201012_DE43A2")
        runner.run(input_data)
    """  # noqa
    return _SklearnRunner(
        tag=tag,
        function_name=function_name,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



import typing as t
import importlib.util
from typing import TYPE_CHECKING

from .lazy_loader import LazyLoader

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

if TYPE_CHECKING:
    import tensorflow as tf
    from tensorflow.python.eager.function import ConcreteFunction
    from tensorflow.python.framework.type_spec import TypeSpec
    from tensorflow.python.training.tracking.base import Trackable
    from tensorflow.python.training.tracking.tracking import AutoTrackable
    from tensorflow.python.saved_model.signature_serialization import (
        _SignatureMap,  # type: ignore[reportPrivateUsage]
    )
    from tensorflow.python.saved_model.function_deserialization import RestoredFunction
else:
    tf = LazyLoader(
        "tf",
        globals(),
        "tensorflow",
        exc_msg=f"tensorflow is required to use {__name__}",
    )

TF_KERAS_DEFAULT_FUNCTIONS = {
    "_default_save_signature",
    "call_and_return_all_conditional_losses",
}

TENSOR_CLASS_NAMES = (
    "RaggedTensor",
    "SparseTensor",
    "TensorArray",
    "EagerTensor",
    "Tensor",
)


def get_tf_version() -> str:
    # courtesy of huggingface/transformers
    _tf_version = ""
    _tf_available = importlib.util.find_spec("tensorflow") is not None
    if _tf_available:
        candidates = (
            "tensorflow",
            "tensorflow-cpu",
            "tensorflow-gpu",
            "tf-nightly",
            "tf-nightly-cpu",
            "tf-nightly-gpu",
            "intel-tensorflow",
            "intel-tensorflow-avx512",
            "tensorflow-rocm",
            "tensorflow-macos",
        )
        # For the metadata, we have to look for both tensorflow and tensorflow-cpu
        for pkg in candidates:
            try:
                _tf_version = importlib_metadata.version(pkg)
                break
            except importlib_metadata.PackageNotFoundError:
                pass
    return _tf_version


def _isinstance_wrapper(
    obj: t.Any, sobj: t.Union[str, type, t.Sequence[t.Any]]
) -> bool:
    """
    `isinstance` wrapper to check tensor spec

    Args:
        obj:
            tensor class to check.
        sobj:
            class used to check with :obj:`obj`. Follows `TENSOR_CLASS_NAME`

    Returns:
        :obj:`bool`
    """
    if not sobj:
        return False
    if isinstance(sobj, str):
        return type(obj).__name__ == sobj.split(".")[-1]
    if isinstance(sobj, (tuple, list, set)):
        return any(_isinstance_wrapper(obj, k) for k in sobj)
    return isinstance(obj, sobj)


def normalize_spec(value: t.Any) -> "TypeSpec":
    """normalize tensor spec"""
    if not _isinstance_wrapper(value, TENSOR_CLASS_NAMES):
        return value

    if _isinstance_wrapper(value, "RaggedTensor"):
        return tf.RaggedTensorSpec.from_value(value)
    if _isinstance_wrapper(value, "SparseTensor"):
        return tf.SparseTensorSpec.from_value(value)
    if _isinstance_wrapper(value, "TensorArray"):
        return tf.TensorArraySpec.from_tensor(value)
    if _isinstance_wrapper(value, ("Tensor", "EagerTensor")):
        return tf.TensorSpec.from_tensor(value)
    return value


def get_input_signatures(
    func: t.Union[t.Any, "RestoredFunction", "ConcreteFunction"]
) -> t.Tuple[t.Any, ...]:
    if hasattr(func, "function_spec"):  # for RestoredFunction
        if func.function_spec.input_signature:
            return ((func.function_spec.input_signature, {}),)
        else:
            return tuple(
                s
                for conc in func.concrete_functions
                for s in get_input_signatures(conc)
            )
    if hasattr(func, "structured_input_signature"):  # for ConcreteFunction
        if func.structured_input_signature is not None:
            return (func.structured_input_signature,)

        if func._arg_keywords is not None:  # TODO(bojiang): using private API
            return (
                (
                    tuple(),
                    {
                        k: normalize_spec(v)
                        for k, v in zip(func._arg_keywords, func.inputs)
                    },
                ),
            )
    return tuple()


def get_arg_names(
    func: t.Union[t.Any, "RestoredFunction", "ConcreteFunction"]
) -> t.Tuple[t.Any, ...]:
    if hasattr(func, "function_spec"):  # for RestoredFunction
        return func.function_spec.arg_names
    if hasattr(func, "structured_input_signature"):  # for ConcreteFunction
        return func._arg_keywords
    return tuple()


def get_output_signature(
    func: t.Union[t.Any, "RestoredFunction", "ConcreteFunction"]
) -> t.Tuple[t.Any, ...]:
    if hasattr(func, "function_spec"):  # for RestoredFunction
        # assume all concrete functions have same signature
        return get_output_signature(func.concrete_functions[0])

    if hasattr(func, "structured_input_signature"):  # for ConcreteFunction
        if func.structured_outputs is not None:
            if isinstance(func.structured_outputs, dict):
                return {
                    k: normalize_spec(v) for k, v in func.structured_outputs.items()
                }
            return func.structured_outputs
        else:
            return tuple(normalize_spec(v) for v in func.outputs)

    return tuple()


def get_restored_functions(m: "Trackable") -> t.Dict[str, t.Any]:
    function_map = {k: getattr(m, k, None) for k in dir(m)}
    return {
        k: v
        for k, v in function_map.items()
        if k not in TF_KERAS_DEFAULT_FUNCTIONS and hasattr(v, "function_spec")
    }


def get_serving_default_function(
    m: "Trackable",
) -> t.Callable[..., t.Any]:
    if not hasattr(m, "signatures"):
        raise EnvironmentError(f"{type(m)} is not a valid SavedModel format.")
    signatures: "_SignatureMap" = m.signatures
    return signatures.get(tf.compat.v2.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY)


def cast_tensor_by_spec(_input: "tf.Tensor", spec: "tf.TensorSpec") -> "tf.Tensor":
    """
    transform dtype & shape following spec
    """
    if not _isinstance_wrapper(spec, "TensorSpec"):
        return _input

    if _isinstance_wrapper(_input, ["Tensor", "EagerTensor"]):
        # TensorFlow issue #43038
        # pylint: disable=unexpected-keyword-arg, no-value-for-parameter
        return tf.cast(_input, dtype=spec.dtype, name=spec.name)
    else:
        return tf.constant(_input, dtype=spec.dtype, name=spec.name)


def _pretty_format_function_call(base: str, name: str, arg_names: t.Tuple[t.Any]):
    if arg_names:
        part_sigs = ", ".join(f"{k}" for k in arg_names)
    else:
        part_sigs = ""

    if name == "__call__":
        return f"{base}({part_sigs})"
    return f"{base}.{name}({part_sigs})"


def _pretty_format_positional(positional: t.List[t.Any]) -> str:
    return f'Positional arguments ({len(positional)} total):\n    * \n{"    * ".join(str(a) for a in positional)}'  # noqa


def pretty_format_function(
    function: t.Callable[..., t.Any], obj: str = "<object>", name: str = "<function>"
) -> str:
    ret = ""
    outs = get_output_signature(function)
    sigs = get_input_signatures(function)
    arg_names = get_arg_names(function)

    if hasattr(function, "function_spec"):
        arg_names = getattr(function, "function_spec").arg_names
    else:
        arg_names = function._arg_keywords  # type: ignore[reportFunctionMemberAccess]

    ret += _pretty_format_function_call(obj, name, arg_names)
    ret += "\n------------\n"

    signature_descriptions = []  # type: t.List[str]

    for index, sig in enumerate(sigs):
        positional, keyword = sig
        signature_descriptions.append(
            "Arguments Option {}:\n  {}\n  Keyword arguments:\n    {}".format(
                index + 1, _pretty_format_positional(positional), keyword
            )
        )

    ret += "\n\n".join(signature_descriptions)
    ret += f"\n\nReturn:\n  {outs}\n\n"
    return ret


def pretty_format_restored_model(model: "AutoTrackable") -> str:
    part_functions = ""

    restored_functions = get_restored_functions(model)
    for name, func in restored_functions.items():
        part_functions += pretty_format_function(func, "model", name)
        part_functions += "\n"

    serving_default = get_serving_default_function(model)
    if serving_default:
        part_functions += pretty_format_function(
            serving_default, "model", "signature['serving_default']"
        )
        part_functions += "\n"

    if not restored_functions and not serving_default:
        return (
            "No serving function was found in the saved model. "
            "In the model implementation, use `tf.function` decorator to mark "
            "the method needed for model serving. \n"
            "Find more details in related TensorFlow docs here "
            "https://www.tensorflow.org/api_docs/python/tf/saved_model/save"
        )
    return f"Found restored functions:\n{part_functions}"



import typing as t
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..utils import LazyLoader
from ..models import Model
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    import numpy as np
    from pandas.core.frame import DataFrame

    from ..models import ModelStore
else:
    np = LazyLoader("np", globals(), "numpy")

try:
    import catboost as cbt
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """catboost is required in order to use module `bentoml.catboost`, install
        catboost with `pip install catboost`. For more information, refers to
        https://catboost.ai/docs/concepts/python-installation.html
        """
    )

_catboost_version = get_pkg_version("catboost")

# TODO: support cbt.Pool runner io container

CATBOOST_EXT = "cbm"


def _get_model_info(
    tag: t.Union[str, Tag],
    model_params: t.Optional[t.Dict[str, t.Union[str, int]]],
    model_store: "ModelStore",
) -> t.Tuple["Model", str, t.Dict[str, t.Any]]:
    model = model_store.get(tag)
    if model.info.module != __name__:
        raise BentoMLException(  # pragma: no cover
            f"Model {tag} was saved with module {model.info.module}, failed loading "
            f"with {__name__}."
        )

    model_file = model.path_of(f"{SAVE_NAMESPACE}.{CATBOOST_EXT}")
    _model_params: t.Dict[str, t.Union[str, int]] = (
        dict() if not model_params else model_params
    )
    for key, value in model.info.options.items():
        if key not in _model_params:
            _model_params[key] = value  # pragma: no cover

    return model, model_file, _model_params


def _load_helper(
    model_file: str, model_params: t.Optional[t.Dict[str, t.Union[str, int]]]
) -> t.Union[
    cbt.core.CatBoost,
    cbt.core.CatBoostClassifier,
    cbt.core.CatBoostRegressor,
]:

    if model_params is not None:
        model_type = model_params["model_type"]

        if model_type == "classifier":
            model = cbt.core.CatBoostClassifier()
        elif model_type == "regressor":
            model = cbt.core.CatBoostRegressor()
        else:
            model = cbt.core.CatBoost()
    else:
        model = cbt.core.CatBoost()

    _m: t.Union[
        cbt.core.CatBoost,
        cbt.core.CatBoostClassifier,
        cbt.core.CatBoostRegressor,
    ] = model.load_model(model_file)
    return _m


@inject
def load(
    tag: t.Union[str, Tag],
    model_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> t.Union[
    cbt.core.CatBoost,
    cbt.core.CatBoostClassifier,
    cbt.core.CatBoostRegressor,
]:

    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        model_params (`t.Dict[str, t.Union[str, Any]]`):
            Parameters for model. Following parameters can be specified:
                - model_type: "classifier" or "regressor" Determine if the model is
                  a `CatBoostClassifier` or `CatBoostRegressor`
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `catboost.core.CatBoostClassifier` or `catboost.core. CatBoostRegressor`
        from BentoML modelstore.

    Examples::
        import bentoml.catboost
        booster = bentoml.catboost.load(
            "my_model:20201012_DE43A2", model_params=dict(model_type="classifier"))
    """  # noqa

    _, _model_file, _model_params = _get_model_info(tag, model_params, model_store)

    return _load_helper(_model_file, _model_params)


@inject
def save(
    name: str,
    model: t.Union[
        cbt.core.CatBoost,
        cbt.core.CatBoostClassifier,
        cbt.core.CatBoostRegressor,
    ],
    *,
    model_params: t.Optional[t.Dict[str, t.Union[str, t.Any]]] = None,
    model_export_parameters: t.Optional[t.Dict[str, t.Any]] = None,
    model_pool: t.Optional["cbt.core.Pool"] = None,
    metadata: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`t.Union[catboost.core.CatBoost, catboost.core.CatBoostClassifier, catboost.CatBoostRegressor]`):
            Instance of model to be saved
        model_params (`t.Dict[str, t.Union[str, t.Any]]`, `optional`, default to `None`):
            Parameters for model. Following parameters can be specified:
                - model_type: "classifier" or "regressor" Determine if the model is
                  a `CatBoostClassifier` or `CatBoostRegressor`
        model_export_parameters (`t.Dict[str, t.Union[str, t.Any]]`, `optional`, default to `None`):
            Export parameters for given model.
        model_pool (`cbt.core.Pool`, `optional`, default to `None`):
            CatBoost data pool for given model.
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        from sklearn.datasets import load_breast_cancer

        import catboost as cbt
        from bentoml.catboost import CatBoostClassifier

        # read in data
        cancer = load_breast_cancer()
        X = cancer.data
        y = cancer.target

        # create and trainmodel
        model = CatBoostClassifier(iterations=2,
                                   depth=2,
                                   learning_rate=1,
                                   loss_function='Logloss',
                                   verbose=True)

        model.fit(X, y)
        ...

        tag = bentoml.catboost.save("my_catboost_model", model, model_params=dict(model_type="classifier"))
        # example tag: my_catboost_model:20211001_AB90F5_153BC4

        # load the booster back:
        loaded = bentoml.catboost.load("my_catboost_model:latest") # or
        loaded = bentoml.catboost.load(tag)

    """  # noqa
    if not model_params:
        model_params = {}

    if "model_type" not in model_params:
        model_params["model_type"] = "classifier"

    context = {
        "framework_name": "catboost",
        "pip_dependencies": [f"catboost=={_catboost_version}"],
    }
    _model = Model.create(
        name,
        module=__name__,
        options=model_params,
        metadata=metadata,
        context=context,
    )

    path = _model.path_of(f"{SAVE_NAMESPACE}.{CATBOOST_EXT}")
    format_ = CATBOOST_EXT
    model.save_model(
        path,
        format=format_,
        export_parameters=model_export_parameters,
        pool=model_pool,
    )

    _model.save(model_store)
    return _model.tag


class _CatBoostRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        model_params: t.Optional[t.Dict[str, t.Union[str, int]]],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        model_info, model_file, _model_params = _get_model_info(
            tag, model_params, model_store
        )
        super().__init__(model_info.tag.name, resource_quota, batch_options)
        self._model_info = model_info
        self._model_file = model_file
        self._predict_fn_name = predict_fn_name
        self._model_params = _model_params

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._model_info.tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        return 1

    @property
    def num_replica(self) -> int:
        return int(round(self.resource_quota.cpu))

    # pylint: disable=attribute-defined-outside-init
    def _setup(self) -> None:
        self._model = _load_helper(self._model_file, self._model_params)
        self._predict_fn = getattr(self._model, self._predict_fn_name)

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        inputs: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", "DataFrame", cbt.Pool],
    ) -> "np.ndarray[t.Any, np.dtype[t.Any]]":
        res = self._predict_fn(inputs)
        return np.asarray(res)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    predict_fn_name: str = "predict",
    *,
    model_params: t.Union[None, t.Dict[str, t.Union[str, int]]] = None,
    resource_quota: t.Union[None, t.Dict[str, t.Any]] = None,
    batch_options: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_CatBoostRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.catboost.load_runner` implements a Runner class that
    wrap around a CatBoost model, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `predict`):
            Options for inference functions. `predict` are the default function.
        model_params (`t.Dict[str, t.Union[str, int]]`, default to `None`):
            Parameters for model. Following parameters can be specified:
                - model_type: "classifier" or "regressor" Determine if the model is
                  a `CatBoostClassifier` or `CatBoostRegressor`
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.catboost` model

    Examples::
        import catboost as cbt
        import bentoml.catboost
        import pandas as pd

        input_data = pd.from_csv("/path/to/csv")
        runner = bentoml.catboost.load_runner("my_model:latest"")
        runner.run(cbt.Pool(input_data))
    """  # noqa
    return _CatBoostRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        model_params=model_params,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



# pylint: disable=redefined-outer-name # pragma: no cover
import os
import sys
import time
import typing as t
import urllib
import logging
import threading
import contextlib
import subprocess
import urllib.request
from typing import TYPE_CHECKING
from contextlib import contextmanager

from .._internal.utils import reserve_free_port
from .._internal.utils import cached_contextmanager

logger = logging.getLogger("bentoml.tests")


if TYPE_CHECKING:
    from aiohttp.typedefs import LooseHeaders
    from starlette.datastructures import Headers
    from starlette.datastructures import FormData

    import docker


clean_context = contextlib.ExitStack()


@cached_contextmanager("{bento}, {config_file}, {workdir}, {docker}, {dev_server}")
def host_bento(
    bento: str,
    config_file: str = "bentoml_config.yml",
    workdir: str = "./",
    docker: bool = False,
    dev_server: bool = False,
) -> t.Generator[str, None, None]:
    # TODO: currently not used

    if not os.path.exists(config_file):
        raise Exception(f"config file not found: {config_file}")

    if docker:
        image = clean_context.enter_context(
            build_api_server_docker_image(bento, "example_service")  # TODO
        )
        host = clean_context.enter_context(
            run_api_server_in_docker(
                image,
                config_file=config_file,
            )
        )
        yield host
    elif dev_server:
        host = clean_context.enter_context(
            run_api_server(
                bento,
                config_file=config_file,
                workdir=workdir,
                dev_server=True,
            )
        )
        yield host
    else:
        host = clean_context.enter_context(
            run_api_server(
                bento,
                config_file=config_file,
                workdir=workdir,
            )
        )
        yield host


async def parse_multipart_form(headers: "Headers", body: bytes) -> "FormData":
    """
    parse starlette forms from headers and body
    """

    from starlette.formparsers import MultiPartParser

    async def async_bytesio(bytes_: bytes) -> t.AsyncGenerator[bytes, None]:
        yield bytes_
        yield b""
        return

    parser = MultiPartParser(headers=headers, stream=async_bytesio(body))
    return await parser.parse()


async def async_request(
    method: str,
    url: str,
    headers: t.Optional["LooseHeaders"] = None,
    data: t.Any = None,
    timeout: t.Optional[int] = None,
) -> t.Tuple[int, "Headers", bytes]:
    """
    raw async request client
    """
    import aiohttp
    from starlette.datastructures import Headers

    async with aiohttp.ClientSession() as sess:
        async with sess.request(
            method, url, data=data, headers=headers, timeout=timeout
        ) as r:
            r_body = await r.read()

    headers = t.cast(t.Mapping[str, str], r.headers)
    return r.status, Headers(headers), r_body


def wait_until_container_ready(container_name, check_message, timeout_seconds=120):
    import docker
    import docker.errors

    docker_client = docker.from_env()

    start_time = time.time()
    while True:
        time.sleep(1)

        # Raise timeout, if exceeds timeout limit
        if time.time() - start_time > timeout_seconds:
            raise TimeoutError(f'Waiting for container "{container_name}" timed out')

        try:
            container_list = docker_client.containers.list(
                filters={"name": container_name}
            )
            if not container_list:
                continue
        except docker.errors.NotFound:
            continue

        logger.info("Container list: " + str(container_list))
        assert (
            len(container_list) == 1
        ), f"should be exact one container with name {container_name}"

        container_log = container_list[0].logs().decode()
        if check_message in container_log:
            logger.info(
                f"Found message indicating container readiness in container log: "
                f"{container_log}"
            )
            break


def _wait_until_api_server_ready(host_url, timeout, container=None, check_interval=1):
    start_time = time.time()
    proxy_handler = urllib.request.ProxyHandler({})
    opener = urllib.request.build_opener(proxy_handler)
    ex = None
    while time.time() - start_time < timeout:
        try:
            if opener.open(f"http://{host_url}/readyz", timeout=1).status == 200:
                return
            elif container and container.status != "running":
                break
            else:
                logger.info("Waiting for host %s to be ready..", host_url)
                time.sleep(check_interval)
        except Exception as e:  # pylint:disable=broad-except
            logger.info(f"retrying to connect to the host {host_url}...")
            ex = e
            time.sleep(check_interval)
        finally:
            if container:
                container_logs = container.logs()
                if container_logs:
                    logger.info(f"Container {container.id} logs:")
                    for log_record in container_logs.decode().split("\r\n"):
                        logger.info(f">>> {log_record}")
    else:
        logger.info("Timeout!")
        raise AssertionError(
            f"Timed out waiting {timeout} seconds for Server {host_url} to be ready, "
            f"exception: {ex}"
        )


@contextmanager
def export_service_bundle(bento_service):
    """
    Export a bentoml service to a temporary directory, yield the path.
    Delete the temporary directory on close.
    """
    import tempfile

    with tempfile.TemporaryDirectory() as path:
        bento_service.save_to_dir(path)
        yield path


@cached_contextmanager("{saved_bundle_path}, {image_tag}")
def build_api_server_docker_image(
    saved_bundle_path, image_tag="test_bentoml_server"
) -> "docker.Image":
    """
    Build the docker image for a saved bentoml bundle, yield the docker image object.
    """

    import docker
    import docker.errors

    client = docker.from_env()
    logger.info(
        f"Building API server docker image from build context: {saved_bundle_path}"
    )
    try:
        image, _ = client.images.build(path=saved_bundle_path, tag=image_tag, rm=False)
        yield image
        client.images.remove(image.id)
    except docker.errors.BuildError as e:
        for line in e.build_log:
            if "stream" in line:
                print(line["stream"].strip())
        raise


@cached_contextmanager("{image.id}")
def run_api_server_in_docker(image, config_file=None, timeout=90):
    """
    Launch a bentoml service container from a docker image, yields the host URL.
    """
    import docker

    client = docker.from_env()

    with reserve_free_port() as port:
        pass

    command_args = "--workers 1"

    if config_file is not None:
        environment = ["BENTOML_CONFIG=/home/bentoml/bentoml_config.yml"]
        volumes = {
            os.path.abspath(config_file): {
                "bind": "/home/bentoml/bentoml_config.yml",
                "mode": "ro",
            }
        }
    else:
        environment = None
        volumes = None

    container = client.containers.run(
        image=image.id,
        command=command_args,
        tty=True,
        ports={"5000/tcp": port},
        detach=True,
        volumes=volumes,
        environment=environment,
    )

    try:
        host_url = f"127.0.0.1:{port}"
        _wait_until_api_server_ready(host_url, timeout, container)
        yield host_url
    finally:
        print(container.logs())
        container.stop()
        container.remove()
        time.sleep(1)  # make sure container stopped & deleted


@contextmanager
def run_api_server(
    bento,
    workdir=None,
    config_file=None,
    dev_server=False,
    timeout=90,
):
    """
    Launch a bentoml service directly by the bentoml CLI, yields the host URL.
    """
    workdir = workdir if workdir is not None else "./"

    serve_cmd = "serve"

    my_env = os.environ.copy()

    with reserve_free_port() as port:
        cmd = [sys.executable, "-m", "bentoml", serve_cmd]

        if not dev_server:
            cmd += ["--production"]

        if port:
            cmd += ["--port", f"{port}"]
        cmd += [bento]
        cmd += ["--working-dir", workdir]

    print(cmd)

    def print_log(p):
        try:
            for line in p.stdout:
                print(line.decode(), end="")
        except ValueError:
            pass

    if config_file is not None:
        my_env["BENTOML_CONFIG"] = os.path.abspath(config_file)

    p = subprocess.Popen(
        cmd,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        env=my_env,
    )
    try:
        threading.Thread(target=print_log, args=(p,), daemon=True).start()
        host_url = f"127.0.0.1:{port}"
        _wait_until_api_server_ready(host_url, timeout=timeout)
        yield host_url
    finally:
        # TODO: can not terminate the subprocess on Windows
        p.terminate()
        p.wait()



from ._internal.frameworks.pytorch_lightning import load
from ._internal.frameworks.pytorch_lightning import save
from ._internal.frameworks.pytorch_lightning import load_runner

__all__ = ["load", "load_runner", "save"]



from ._internal.frameworks.transformers import load
from ._internal.frameworks.transformers import save
from ._internal.frameworks.transformers import load_runner
from ._internal.frameworks.transformers import import_from_huggingface_hub

__all__ = ["import_from_huggingface_hub", "load", "load_runner", "save"]



import os
import shutil
import typing as t
import logging
from functools import partial

logger = logging.getLogger(__name__)


class PrometheusClient:
    def __init__(
        self,
        *,
        namespace: str = "",
        multiproc: bool = True,
        multiproc_dir: t.Optional[str] = None,
    ):
        """
        Set up multiproc_dir for prometheus to work in multiprocess mode,
        which is required when working with Gunicorn server

        Warning: for this to work, prometheus_client library must be imported after
        this function is called. It relies on the os.environ['prometheus_multiproc_dir']
        to properly setup for multiprocess mode
        """
        self.multiproc = multiproc
        self.namespace = namespace
        self._registry = None

        if multiproc:
            assert multiproc_dir is not None, "multiproc_dir must be provided"
            os.environ["prometheus_multiproc_dir"] = multiproc_dir

    @classmethod
    def clean_multiproc_dir(cls, multiproc_dir: t.Optional[str]):
        assert multiproc_dir is not None, "multiproc_dir must be provided"
        logger.debug("Setting up prometheus_multiproc_dir: %s", multiproc_dir)
        # Wipe prometheus metrics directory between runs
        # https://github.com/prometheus/client_python#multiprocess-mode-gunicorn
        # Ignore errors so it does not fail when directory does not exist
        shutil.rmtree(multiproc_dir, ignore_errors=True)
        os.makedirs(multiproc_dir, exist_ok=True)

    @property
    def registry(self):
        if self._registry is None:
            from prometheus_client import multiprocess
            from prometheus_client import CollectorRegistry

            registry = CollectorRegistry()
            if self.multiproc:
                multiprocess.MultiProcessCollector(registry)
            self._registry = registry
        return self._registry

    # review
    # @classmethod
    # def mark_process_dead(cls, pid: int):
    #     # TODO(jiang)
    #     pass

    @staticmethod
    def mark_process_dead(pid: int) -> None:
        from prometheus_client import multiprocess

        multiprocess.mark_process_dead(pid)

    def start_http_server(self, port: int, addr: str = "") -> None:
        from prometheus_client import start_http_server

        start_http_server(port=port, addr=addr, registry=self.registry)

    def generate_latest(self):
        from prometheus_client import generate_latest

        return generate_latest(self.registry)

    @property
    def CONTENT_TYPE_LATEST(self) -> str:
        from prometheus_client import CONTENT_TYPE_LATEST

        return CONTENT_TYPE_LATEST

    @property
    def Histogram(self):
        from prometheus_client import Histogram as Operator

        return partial(Operator, namespace=self.namespace, registry=self.registry)

    @property
    def Counter(self):
        from prometheus_client import Counter as Operator

        return partial(Operator, namespace=self.namespace, registry=self.registry)

    @property
    def Summary(self):
        from prometheus_client import Summary as Operator

        return partial(Operator, namespace=self.namespace, registry=self.registry)

    @property
    def Gauge(self):
        from prometheus_client import Gauge as Operator

        return partial(Operator, namespace=self.namespace, registry=self.registry)



import os
import typing as t
from typing import TYPE_CHECKING

import numpy as np
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import Model
from ..models import PTH_EXT
from ..models import YAML_EXT
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..runner.utils import Params
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ..models import ModelStore
try:
    # pylint: disable=unused-import;
    import torch
    import detectron2  # noqa F401
    import detectron2.config as config
    import detectron2.modeling as modeling
    import detectron2.checkpoint as checkpoint
    from detectron2.checkpoint import DetectionCheckpointer


except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """detectron2 is required in order to use module `bentoml.detectron`,
        install detectron2 with `pip install detectron2`. For more
        information, refers to
        https://detectron2.readthedocs.io/en/latest/tutorials/install.html
        """
    )

_detectron2_version = get_pkg_version("detectron2")
_torch_version = get_pkg_version("torch")


@inject
def load(
    tag: t.Union[str, Tag],
    device: str = "cpu",
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "torch.nn.Module":
    """
    Load a model from BentoML local modelstore with given tag.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        device (`str`, `optional`, default to ``cpu``):
            Device type to cast model. Default behaviour similar
             to :obj:`torch.device("cuda")` Options: "cuda" or "cpu".
             If None is specified then return default config.MODEL.DEVICE
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `torch.nn.Module`

    Examples::
        import bentoml.detectron
        model = bentoml.detectron.load(
            "my_detectron_model:20201012_DE43A2")
    """  # noqa: LN001

    model_info = model_store.get(tag)
    if model_info.info.module != __name__:  # pragma: no cover
        raise BentoMLException(
            f"Model {tag} was saved with"
            f" module {model_info.info.module},"
            f" failed loading with {__name__}."
        )

    cfg: config.CfgNode = config.get_cfg()

    weight_path = model_info.path_of(f"{SAVE_NAMESPACE}{PTH_EXT}")
    yaml_path = model_info.path_of(f"{SAVE_NAMESPACE}{YAML_EXT}")

    if os.path.isfile(yaml_path):
        cfg.merge_from_file(yaml_path)

    if device:
        cfg.MODEL.DEVICE = device

    model: "torch.nn.Module" = modeling.build_model(cfg)
    if device:
        model.to(device)

    model.eval()

    checkpointer: "DetectionCheckpointer" = checkpoint.DetectionCheckpointer(model)

    checkpointer.load(weight_path)
    return model


@inject
def save(
    name: str,
    model: "torch.nn.Module",
    *,
    model_config: t.Optional[config.CfgNode] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`torch.nn.Module`):
            Instance of detectron2 model to be saved.
        model_config (`detectron2.config.CfgNode`, `optional`, default to `None`):
            model config from :meth:`detectron2.model_zoo.get_config_file`
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        import bentoml.detectron

        # import some common detectron2 utilities
        import detectron2
        from detectron2 import model_zoo
        from detectron2.config import get_cfg
        from detectron2.modeling import build_model

        model_url: str = "COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml"
        cfg: "CfgNode" = get_cfg()
        cfg.merge_from_file(model_zoo.get_config_file(model_url))
        # set threshold for this model
        cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.5
        cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(model_url)
        cloned = cfg.clone()
        cloned.MODEL.DEVICE = "cpu"
        model: torch.nn.Module = build_model(cloned)

        tag = bentoml.detectron.save(
            "my_detectron_model",
            model,
            model_config=cfg,
        )

        # example tag: my_detectron_model:20211018_B9EABF

        # load the model back:
        loaded = bentoml.detectron.load("my_detectron_model:latest") # or
        loaded = bentoml.detectron.load(tag)

    """  # noqa

    context: t.Dict[str, t.Any] = {
        "framework_name": "detectron2",
        "pip_dependencies": [
            f"detectron2=={_detectron2_version}",
            f"torch=={_torch_version}",
        ],
    }
    options: t.Dict[str, t.Any] = dict()

    _model = Model.create(
        name,
        module=__name__,
        options=options,
        context=context,
        metadata=metadata,
    )

    checkpointer = checkpoint.DetectionCheckpointer(model, save_dir=_model.path)
    checkpointer.save(SAVE_NAMESPACE)
    if model_config:
        with open(
            _model.path_of(f"{SAVE_NAMESPACE}{YAML_EXT}"),
            "w",
            encoding="utf-8",
        ) as ouf:
            ouf.write(model_config.dump())

    _model.save(model_store)

    return _model.tag


class _DetectronRunner(Runner):
    @inject
    # TODO add partial_kwargs @larme
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(tag, resource_quota, batch_options)
        self._tag = tag
        self._predict_fn_name = predict_fn_name
        self._model_store = model_store

    @property
    def required_models(self) -> t.List[str]:
        return [self._tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        return 1

    @property
    def num_replica(self) -> int:
        if self.resource_quota.on_gpu:
            return len(self.resource_quota.gpus)
        return 1

    # pylint: disable=attribute-defined-outside-init
    def _setup(self) -> None:
        if self.resource_quota.on_gpu:
            device = "cuda"
        else:
            device = "cpu"
        self._model = load(self._tag, device, self._model_store)
        self._predict_fn = getattr(self._model, self._predict_fn_name)

    def _run_batch(
        self,
        *args: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor],
    ) -> "np.ndarray[t.Any, np.dtype[t.Any]]":
        params = Params[t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor]](
            *args
        )

        def _mapping(
            item: t.Union["np.ndarray[t.Any, np.dtype[t.Any]]", torch.Tensor]
        ) -> torch.Tensor:
            if isinstance(item, np.ndarray):
                return torch.from_numpy(item)
            return item

        params = params.map(_mapping)

        inputs = [{"image": image} for image in params.args]

        res = self._predict_fn(inputs)
        return np.asarray(res, dtype=object)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    predict_fn_name: str = "__call__",
    *,
    resource_quota: t.Union[None, t.Dict[str, t.Any]] = None,
    batch_options: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> _DetectronRunner:
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.detectron.load_runner` implements a Runner class that
    wrap around a torch.nn.Module model, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `__call__`):
            Options for inference functions. Default to `__call__`
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.detectron` model

    Examples:
        TODO
    """  # noqa
    return _DetectronRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



import sys
import typing as t
import logging
from typing import TYPE_CHECKING

from ..types import Tag
from ..runner import Runner
from ...exceptions import BentoMLException
from ..bento.bento import _get_default_bento_readme
from .inference_api import InferenceAPI
from ..io_descriptors import IODescriptor
from ..utils.validation import validate_tag_str

if TYPE_CHECKING:
    from starlette.types import ASGIApp
    from starlette.middleware import Middleware
    from starlette.applications import Starlette

    from ..bento import Bento


WSGI_APP = t.Callable[
    [t.Callable[..., t.Any], t.Mapping[str, t.Any]], t.Iterable[bytes]
]

logger = logging.getLogger(__name__)


class Service:
    """The service definition is the manifestation of the Service Oriented Architecture
    and the core building block in BentoML where users define the service runtime
    architecture and model serving logic.

    A BentoML service is defined via instantiate this Service class. When creating a
    Service instance, user must provide a Service name and list of runners that are
    required by this Service. The instance can then be used to define InferenceAPIs via
    the `api` decorator.
    """

    _apis: t.Dict[str, InferenceAPI] = {}

    # Name of the service, it is a required parameter for __init__
    name: str
    # Tag/Bento/Version are only applicable if the service was load from a bento
    tag: t.Optional[Tag] = None
    bento: t.Optional["Bento"] = None
    version: t.Optional[str] = None
    # Working dir of the service, set when the service was load from a bento
    _working_dir: t.Optional[str] = None
    # Import path set by .loader.import_service method
    _import_str: t.Optional[str] = None
    # For docs property
    _doc: t.Optional[str] = None

    def __init__(self, name: str, runners: t.Optional[t.List[Runner]] = None):
        lower_name = name.lower()

        if name != lower_name:
            logger.warning(f"converting {name} to lowercase: {lower_name}")

        # Service name must be a valid dns1123 subdomain string
        validate_tag_str(lower_name)
        self.name = lower_name

        if runners is not None:
            assert all(
                isinstance(r, Runner) for r in runners
            ), "Service runners list must only contain runner instances"
            self.runners = {r.name: r for r in runners}
        else:
            self.runners: t.Dict[str, Runner] = {}

        self._mount_apps: t.List[t.Tuple[t.Union["ASGIApp", WSGI_APP], str, str]] = []
        self._middlewares: t.List[t.Tuple[t.Type["Middleware"], t.Any]] = []

    def _on_asgi_app_startup(self) -> None:
        # TODO: initialize Local Runner instances or Runner Clients here
        # TODO(P1): add `@svc.on_startup` decorator for adding user-defined hook
        pass

    def _on_asgi_app_shutdown(self) -> None:
        # TODO(P1): add `@svc.on_shutdown` decorator for adding user-defined hook
        pass

    def __del__(self):
        # working dir was added to sys.path in the .loader.import_service function
        if self._working_dir and sys.path:
            sys.path.remove(self._working_dir)

    def api(
        self,
        input: IODescriptor[t.Any],  # noqa # pylint: disable=redefined-builtin
        output: IODescriptor[t.Any],
        name: t.Optional[str] = None,
        doc: t.Optional[str] = None,
        route: t.Optional[str] = None,
    ) -> t.Callable[[t.Callable[..., t.Any]], t.Callable[..., t.Any]]:
        """Decorator for adding InferenceAPI to this service"""

        D = t.TypeVar("D", bound=t.Callable[..., t.Any])

        def decorator(func: D) -> D:
            from ..io_descriptors import Multipart

            if isinstance(output, Multipart):
                logger.warning(
                    f"Found Multipart as the output of API `{name or func.__name__}`. "
                    "Multipart response is rarely used in the real world,"
                    " less clients/browsers support it. "
                    "Make sure you know what you are doing."
                )
            self._add_inference_api(func, input, output, name, doc, route)
            return func

        return decorator

    def _add_inference_api(
        self,
        func: t.Callable[..., t.Any],
        input: IODescriptor[t.Any],  # noqa # pylint: disable=redefined-builtin
        output: IODescriptor[t.Any],
        name: t.Optional[str],
        doc: t.Optional[str],
        route: t.Optional[str],
    ) -> None:
        api = InferenceAPI(
            name=name,
            user_defined_callback=func,
            input_descriptor=input,
            output_descriptor=output,
            doc=doc,
            route=route,
        )

        if api.name in self._apis:
            raise BentoMLException(
                f"API {api.name} is already defined in Service {self.name}"
            )
        self._apis[api.name] = api

    @property
    def asgi_app(self) -> "Starlette":
        logger.info("Initializing HTTP server for %s", self)
        from ..server.service_app import ServiceAppFactory

        return ServiceAppFactory(self)()

    def mount_asgi_app(
        self, app: "ASGIApp", path: str = "/", name: t.Optional[str] = None
    ) -> None:
        self._mount_apps.append((app, path, name))  # type: ignore

    def mount_wsgi_app(
        self, app: WSGI_APP, path: str = "/", name: t.Optional[str] = None
    ) -> None:
        from starlette.middleware.wsgi import WSGIMiddleware

        self._mount_apps.append((WSGIMiddleware(app), path, name))  # type: ignore

    def add_asgi_middleware(
        self, middleware_cls: t.Type["Middleware"], **options: t.Any
    ) -> None:
        self._middlewares.append((middleware_cls, options))

    def openapi_doc(self):
        from .openapi import get_service_openapi_doc

        return get_service_openapi_doc(self)

    def __str__(self):
        if self.bento:
            return f'bentoml.Service(tag="{self.tag}", ' f'path="{self.bento.path}")'
        elif self._import_str and self._working_dir:
            return (
                f'bentoml.Service(name="{self.name}", '
                f'import_str="{self._import_str}", '
                f'working_dir="{self._working_dir}")'
            )
        else:
            return (
                f'bentoml.Service(name="{self.name}", '
                f'runners=[{",".join(self.runners.keys())}])'
            )

    def __repr__(self):
        return self.__str__()

    @property
    def doc(self) -> str:
        if self.bento is not None:
            return self.bento.doc

        return _get_default_bento_readme(self)



import click

from bentoml.bentos import containerize as containerize_bento

from ..utils.docker import validate_tag


def add_containerize_command(cli):
    @cli.command(
        help="Containerizes given Bento into a ready-to-use Docker image.",
    )
    @click.argument("bento_tag", type=click.STRING)
    @click.option(
        "-t",
        "--docker-image-tag",
        help="Docker image tag, default to same as the Bento tag",
        required=False,
        callback=validate_tag,
    )
    @click.option("--build-arg", multiple=True, help="docker image build args")
    @click.option("--label", multiple=True, help="docker image label")
    @click.option("--no-cache", is_flag=True, default=False)
    @click.option("--platform", default=None)
    def containerize(bento_tag, docker_image_tag, build_arg, label, no_cache, platform):
        """Containerize specified Bento.

        BENTO is the target BentoService to be containerized, referenced by its name
        and version in format of name:version. For example: "iris_classifier:v1.2.0"

        `bentoml containerize` command also supports the use of the `latest` tag
        which will automatically use the last built version of your Bento.

        You can provide a tag for the image built by Bento using the
        `--tag` flag. Additionally, you can provide a `--push` flag,
        which will push the built image to the Docker repository specified by the
        image tag.

        You can also prefixing the tag with a hostname for the repository you wish
        to push to.
        e.g. `bentoml containerize IrisClassifier:latest --push --tag
        repo-address.com:username/iris` would build a Docker image called
        `username/iris:latest` and push that to docker repository at repo-address.com.

        By default, the `containerize` command will use the current credentials
        provided by Docker daemon.
        """
        labels = {}
        if label:
            for label_str in label:
                key, value = label_str.split("=")
                labels[key] = value
        build_args = {}
        if build_arg:
            for build_arg_str in build_arg:
                key, value = build_arg_str.split("=")
                build_args[key] = value

        return containerize_bento(
            bento_tag,
            docker_image_tag=docker_image_tag,
            build_args=build_args,
            labels=labels,
            no_cache=no_cache,
            platform=platform,
        )



from ._internal.frameworks.pytorch import load
from ._internal.frameworks.pytorch import save
from ._internal.frameworks.pytorch import load_runner
from ._internal.frameworks.pytorch import PytorchTensorContainer

__all__ = ["PytorchTensorContainer", "load", "load_runner", "save"]



import sys
import json
import typing as t
import dataclasses
from typing import TYPE_CHECKING

from starlette.requests import Request
from starlette.responses import Response

from .base import JSONType
from .base import IODescriptor
from ...exceptions import BadInput
from ...exceptions import MissingDependencyException
from ..utils.lazy_loader import LazyLoader

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    import pydantic
    from pydantic import BaseModel
    from pandas.core.frame import DataFrame  # noqa: F401
else:  # pragma: no cover
    np = LazyLoader("np", globals(), "numpy")
    pd = LazyLoader("pd", globals(), "pandas")
    pydantic = LazyLoader("pydantic", globals(), "pydantic")


MIME_TYPE_JSON = "application/json"

_SerializableObj = t.Union[
    "np.ndarray[t.Any, np.dtype[t.Any]]", t.Type["BaseModel"], "DataFrame", t.Any
]


class DefaultJsonEncoder(json.JSONEncoder):  # pragma: no cover
    def default(self, o: _SerializableObj) -> t.Any:
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)

        if "numpy" in sys.modules:
            if isinstance(o, np.generic):
                return o.item()

            if isinstance(o, np.ndarray):
                return o.tolist()

        if "pandas" in sys.modules:
            if isinstance(o, (pd.DataFrame, pd.Series)):
                return o.to_dict()  # type: ignore

        if "pydantic" in sys.modules:
            if pydantic is not None and isinstance(o, pydantic.BaseModel):
                obj_dict = o.dict()
                if "__root__" in obj_dict:
                    obj_dict = obj_dict.get("__root__")
                return obj_dict

        return super().default(o)


class JSON(IODescriptor[JSONType]):
    """
    `JSON` defines API specification for the inputs/outputs of a Service, where either
     inputs will be converted to or outputs will be converted from a JSON representation
     as specified in your API function signature.

    .. Toy implementation of a sklearn service::
        # sklearn_svc.py
        import pandas as pd
        import numpy as np
        from bentoml.io import PandasDataFrame, JSON
        import bentoml.sklearn

        input_spec = PandasDataFrame.from_sample(pd.DataFrame(np.array([[5,4,3,2]])))

        runner = bentoml.sklearn.load_runner("sklearn_model_clf")

        svc = bentoml.Service("iris-classifier", runners=[runner])

        @svc.api(input=input_spec, output=JSON())
        def predict(input_arr: pd.DataFrame):
            res = runner.run_batch(input_arr)  # type: np.ndarray
            return {"res":pd.DataFrame(res).to_json(orient='record')}

    Users then can then serve this service with `bentoml serve`::
        % bentoml serve ./sklearn_svc.py:svc --auto-reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "iris-classifier" defined in "sklearn_svc.py"
        [INFO] API Server running on http://0.0.0.0:5000

    Users can then send a cURL requests like shown in different terminal session::
        % curl -X POST -H "Content-Type: application/json" --data
          '[{"0":5,"1":4,"2":3,"3":2}]' http://0.0.0.0:5000/predict

        {"res":"[{\"0\":1}]"}%

    Args:
        pydantic_model (`pydantic.BaseModel`, `optional`, default to `None`):
            Pydantic model schema.
        validate_json (`bool`, `optional`, default to `True`):
            If True, then use Pydantic model specified above to validate given JSON.
        json_encoder (`Type[json.JSONEncoder]`, default to
                      `~bentoml._internal.io_descriptor.json.DefaultJsonEncoder`):
            JSON encoder.
    Returns:
        IO Descriptor that in JSON format.
    """

    def __init__(
        self,
        pydantic_model: t.Optional["BaseModel"] = None,
        validate_json: bool = True,
        json_encoder: t.Type[json.JSONEncoder] = DefaultJsonEncoder,
    ):
        if pydantic_model is not None:  # pragma: no cover
            if pydantic is None:
                raise MissingDependencyException(
                    "`pydantic` must be installed to use `pydantic_model`"
                )
            self._pydantic_model = pydantic_model

        self._validate_json = validate_json
        self._json_encoder = json_encoder

    def openapi_schema_type(self) -> t.Dict[str, t.Any]:
        if hasattr(self, "_pydantic_model"):
            return self._pydantic_model.schema()
        return {"type": "object"}

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {MIME_TYPE_JSON: {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {MIME_TYPE_JSON: {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> JSONType:
        json_obj = await request.json()
        if hasattr(self, "_pydantic_model") and self._validate_json:
            if pydantic is None:
                raise MissingDependencyException(
                    "`pydantic` must be installed to use `pydantic_model`"
                )
            try:
                return self._pydantic_model.parse_obj(json_obj)
            except pydantic.ValidationError:
                raise BadInput("Invalid JSON Request received")
        return json_obj

    async def to_http_response(self, obj: JSONType) -> Response:
        json_str = json.dumps(
            obj,
            cls=self._json_encoder,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
        )
        return Response(json_str, media_type=MIME_TYPE_JSON)



import typing as t
from typing import TYPE_CHECKING

from starlette.requests import Request
from multipart.multipart import parse_options_header
from starlette.responses import Response

from .base import IOType
from .base import IODescriptor
from ...exceptions import InvalidArgument
from ...exceptions import BentoMLException
from ..utils.formparser import populate_multipart_requests
from ..utils.formparser import concat_to_multipart_responses

if TYPE_CHECKING:
    # noqa: F811
    from .file import File
    from .json import JSON
    from .text import Text
    from .image import Image
    from .numpy import NumpyNdarray
    from .pandas import PandasSeries
    from .pandas import PandasDataFrame


MultipartIO = t.Dict[str, IOType]


class Multipart(IODescriptor[MultipartIO]):
    """
    `Multipart` defines API specification for the inputs/outputs of a Service, where inputs/outputs
     of a Service can receive/send a `multipart` request/responses as specified in your API function signature.

    .. Toy implementation of a sklearn service::
        # sklearn_svc.py
        import bentoml
        from bentoml.io import NumpyNdarray, Multipart, JSON
        import bentoml.sklearn

        runner = bentoml.sklearn.load_runner("sklearn_model_clf")

        svc = bentoml.Service("iris-classifier", runners=[runner])
        input_spec = Multipart(arr=NumpyNdarray(), annotations=JSON())
        output_spec = Multipart(output=NumpyNdarray(), result=JSON())

        @svc.api(input=input_spec, output=output_spec)
        def predict(arr, annotations):
            res = runner.run(arr)
            return {"output":res, "result":annotations}

    Users then can then serve this service with `bentoml serve`::
        % bentoml serve ./sklearn_svc.py:svc --reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "iris-classifier" defined in "sklearn_svc.py"
        [INFO] API Server running on http://0.0.0.0:5000

    Users can then send a cURL requests like shown in different terminal session::
        % curl -X POST -H "Content-Type: multipart/form-data" -F annotations=@test.json -F arr='[5,4,3,2]' http://0.0.0.0:5000/predict

        --b1d72c201a064ecd92a17a412eb9208e
        Content-Disposition: form-data; name="output"
        content-length: 1
        content-type: application/json

        1
        --b1d72c201a064ecd92a17a412eb9208e
        Content-Disposition: form-data; name="result"
        content-length: 13
        content-type: application/json

        {"foo":"bar"}
        --b1d72c201a064ecd92a17a412eb9208e--

    Args:
        inputs (`Dict[str, IODescriptor]`):
            Dictionary consisting keys as inputs definition for a Multipart request/response, values
             as IODescriptor supported by BentoML. Currently we support Image, NumpyNdarray, PandasDataFrame,
             PandasSeries, Text, File.

    .. notes::
        Make sure to match your input params in your API function to the keys defined under `Multipart`::
            ┌───────────────────────────────────────────────────────┐
            │                                                       │
            │ ┌───────────────────────────────────────────────────┐ │
            │ │ Multipart(arr=NumpyNdarray(), annotations=JSON()) │ │
            │ └─────────────────────┬─────────────┬───────────────┘ │
            │                       │             │                 │
            │                       │       ┌─────┘                 │
            │                       │       │                       │
            │       ┌───────────────▼───────▼─────────┐             │
            │       │  def predict(arr, annotations): │             │
            │       └─────────────────────────────────┘             │
            │                                                       │
            └───────────────────────────────────────────────────────┘

    Returns:
        IO Descriptor that represents Multipart request/response.
    """  # noqa: LN001

    def __init__(
        self,
        **inputs: t.Union[
            "Image",
            "JSON",
            "Text",
            "NumpyNdarray",
            "PandasDataFrame",
            "PandasSeries",
            "File",
        ],
    ):
        for descriptor in inputs.values():
            if isinstance(descriptor, Multipart):  # pragma: no cover
                raise InvalidArgument(
                    "Multipart IO can not contain nested Multipart item"
                )
        self._inputs: t.Dict[
            str,
            t.Union[
                "Image",
                "JSON",
                "Text",
                "NumpyNdarray",
                "PandasDataFrame",
                "PandasSeries",
                "File",
            ],
        ] = inputs

    def openapi_schema_type(self) -> t.Dict[str, t.Any]:
        return {
            "type": "object",
            "properties": {
                k: io.openapi_schema_type() for k, io in self._inputs.items()
            },
        }

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {"multipart/form-data": {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {"multipart/form-data": {"schema": self.openapi_schema_type()}}

    async def from_http_request(self, request: Request) -> MultipartIO:
        ctype, _ = parse_options_header(request.headers["content-type"])
        if ctype != b"multipart/form-data":
            raise BentoMLException(
                f"{self.__class__.__name__} only accepts `multipart/form-data` as Content-Type header, got {ctype} instead."
            )

        res: MultipartIO = dict()
        reqs = await populate_multipart_requests(request)

        for k, i in self._inputs.items():
            req = reqs[k]
            v = await i.from_http_request(req)
            res[k] = v
        return res

    async def to_http_response(self, obj: MultipartIO) -> Response:
        res_mapping: t.Dict[str, Response] = {}
        for k, io_ in self._inputs.items():
            data = obj[k]
            # TODO(aarnphm): fix with stubs
            res_mapping[k] = await io_.to_http_response(data)  # type: ignore[reportGeneralTypeIssue]
        return await concat_to_multipart_responses(res_mapping)



import typing as t
from typing import TYPE_CHECKING

import numpy as np
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import Model
from ..models import JSON_EXT
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    import pandas as pd

    from ..models import ModelStore

try:
    import xgboost as xgb
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """xgboost is required in order to use module `bentoml.xgboost`, install
        xgboost with `pip install xgboost`. For more information, refers to
        https://xgboost.readthedocs.io/en/latest/install.html
        """
    )

_xgboost_version = get_pkg_version("xgboost")

# TODO: support xgb.DMatrix runner io container
# from bentoml.runner import RunnerIOContainer, register_io_container
# class DMatrixContainer(RunnerIOContainer):
#     batch_type = xgb.DMatrix
#     item_type = xgb.DMatrix
#
#     def flatten(self):
#         pass
#
#     def squeeze(self):
#         pass
#
#     def serialize(self):
#         pass
#
#     def deserialize(self):
#         pass
#
# register_io_container(DMatrixContainer)


def _get_model_info(
    tag: t.Union[str, Tag],
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]],
    model_store: "ModelStore",
) -> t.Tuple["Model", str, t.Dict[str, t.Any]]:
    model = model_store.get(tag)
    if model.info.module != __name__:
        raise BentoMLException(  # pragma: no cover
            f"Model {tag} was saved with"
            f" module {model.info.module},"
            f" failed loading with {__name__}."
        )
    model_file = model.path_of(f"{SAVE_NAMESPACE}{JSON_EXT}")
    _booster_params = dict() if not booster_params else booster_params
    for key, value in model.info.options.items():
        if key not in _booster_params:
            _booster_params[key] = value  # pragma: no cover
    if "nthread" not in _booster_params:
        _booster_params["nthread"] = -1  # apply default nthread parameter

    return model, model_file, _booster_params


@inject
def load(
    tag: t.Union[str, Tag],
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "xgb.core.Booster":
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (`str`):
            Tag of a saved model in BentoML local modelstore.
        booster_params (`t.Dict[str, t.Union[str, int]]`):
            Params for xgb.core.Booster initialization
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        an instance of `xgboost.core.Booster` from BentoML modelstore.

    Examples::
        import bentoml.xgboost
        booster = bentoml.xgboost.load(
            'my_model:20201012_DE43A2', booster_params=dict(gpu_id=0))
    """  # noqa
    _, _model_file, _booster_params = _get_model_info(tag, booster_params, model_store)

    return xgb.core.Booster(
        params=_booster_params,
        model_file=_model_file,
    )


@inject
def save(
    name: str,
    model: "xgb.core.Booster",
    *,
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`xgboost.core.Booster`):
            Instance of model to be saved
        booster_params (`t.Dict[str, t.Union[str, int]]`):
            Params for booster initialization
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        import xgboost as xgb
        import bentoml.xgboost

        # read in data
        dtrain = xgb.DMatrix('demo/data/agaricus.txt.train')
        dtest = xgb.DMatrix('demo/data/agaricus.txt.test')
        # specify parameters via map
        param = dict(max_depth=2, eta=1, objective='binary:logistic')
        num_round = 2
        bst = xgb.train(param, dtrain, num_round)
        ...

        tag = bentoml.xgboost.save("my_xgboost_model", bst, booster_params=param)
        # example tag: my_xgboost_model:20210929_153BC4

        # load the booster back:
        bst = bentoml.xgboost.load("my_xgboost_model:latest") # or
        bst = bentoml.xgboost.load(tag)
    """  # noqa
    context: t.Dict[str, t.Any] = {
        "framework_name": "xgboost",
        "pip_dependencies": [f"xgboost=={_xgboost_version}"],
    }

    _model = Model.create(
        name,
        module=__name__,
        options=booster_params,
        context=context,
        metadata=metadata,
    )

    model.save_model(_model.path_of(f"{SAVE_NAMESPACE}{JSON_EXT}"))
    _model.save(model_store)

    return _model.tag


class _XgBoostRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        predict_fn_name: str,
        booster_params: t.Optional[t.Dict[str, t.Union[str, int]]],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        super().__init__(str(tag), resource_quota, batch_options)
        model_info, model_file, booster_params = _get_model_info(
            tag, booster_params, model_store
        )

        self._model_store = model_store
        self._model_info = model_info
        self._model_file = model_file
        self._predict_fn_name = predict_fn_name
        booster_params = dict() if booster_params is None else booster_params
        self._booster_params = self._setup_booster_params(booster_params)

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._model_info.tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        if self.resource_quota.on_gpu:
            return 1
        nthreads = self._booster_params.get("nthread", -1)  # type: int
        if nthreads == -1:
            return int(round(self.resource_quota.cpu))
        return nthreads

    @property
    def num_replica(self) -> int:
        if self.resource_quota.on_gpu:
            return len(self.resource_quota.gpus)
        return 1

    def _setup_booster_params(
        self, booster_params: t.Dict[str, t.Any]
    ) -> t.Dict[str, t.Any]:
        if self.resource_quota.on_gpu:
            booster_params["predictor"] = "gpu_predictor"
            booster_params["tree_method"] = "gpu_hist"
            # Use the first device reported by CUDA runtime
            booster_params["gpu_id"] = 0
            booster_params["nthread"] = 1
        else:
            booster_params["predictor"] = "cpu_predictor"
            booster_params["nthread"] = int(round(self.resource_quota.cpu))

        return booster_params

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:  # type: ignore[override]
        self._model = xgb.core.Booster(
            params=self._booster_params,
            model_file=self._model_file,
        )
        self._predict_fn = getattr(self._model, self._predict_fn_name)

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[override]
        self, input_data: t.Union[np.ndarray, "pd.DataFrame", xgb.DMatrix]
    ) -> "np.ndarray":
        if not isinstance(input_data, xgb.DMatrix):
            input_data = xgb.DMatrix(input_data)
        res = self._predict_fn(input_data)
        return np.asarray(res)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    predict_fn_name: str = "predict",
    *,
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
    batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_XgBoostRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.xgboost.load_runner` implements a Runner class that
    wrap around a Xgboost booster model, which optimize it for the BentoML runtime.

    Args:
        tag (`str`):
            Model tag to retrieve model from modelstore
        predict_fn_name (`str`, default to `predict`):
            Options for inference functions. If you want to use `run`
             or `run_batch` in a thread context then use `inplace_predict`.
             Otherwise, `predict` are the de facto functions.
        booster_params (`t.Dict[str, t.Union[str, int]]`, default to `None`):
            Parameters for boosters. Refers to https://xgboost.readthedocs.io/en/latest/parameter.html
             for more information
        resource_quota (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure resources allocation for runner.
        batch_options (`t.Dict[str, t.Any]`, default to `None`):
            Dictionary to configure batch options for runner in a service context.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        Runner instances for `bentoml.xgboost` model

    Examples::
        import xgboost as xgb
        import bentoml.xgboost
        import pandas as pd

        input_data = pd.from_csv("/path/to/csv")
        runner = bentoml.xgboost.load_runner("my_model:20201012_DE43A2")
        runner.run(xgb.DMatrix(input_data))
    """  # noqa
    return _XgBoostRunner(
        tag=tag,
        predict_fn_name=predict_fn_name,
        booster_params=booster_params,
        resource_quota=resource_quota,
        batch_options=batch_options,
        model_store=model_store,
    )



from ._internal.frameworks.keras import load
from ._internal.frameworks.keras import save
from ._internal.frameworks.keras import get_session
from ._internal.frameworks.keras import load_runner

__all__ = ["load", "load_runner", "save", "get_session"]



from ._internal.frameworks.statsmodels import load
from ._internal.frameworks.statsmodels import save
from ._internal.frameworks.statsmodels import load_runner

__all__ = ["load", "load_runner", "save"]



import os
import sys
import typing as t
import logging
import importlib
from typing import TYPE_CHECKING
from pathlib import Path
from functools import partial
from distutils.dir_util import copy_tree

import yaml
from simple_di import inject
from simple_di import Provide

from bentoml import Tag
from bentoml import Runner
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..utils import LazyLoader
from ..models import Model
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from ..bento.pip_pkg import split_requirement
from ..bento.pip_pkg import packages_distributions
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:  # pragma: no cover
    from spacy.vocab import Vocab
    from thinc.config import Config
    from spacy.tokens.doc import Doc

    from ..models import ModelStore

if sys.version_info >= (3, 8):
    from typing import Literal
else:  # pragma: no cover
    from typing_extensions import Literal

try:
    import spacy
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """\
        `spacy` is required to use with `bentoml.spacy`.
        Instruction: Refers to https://spacy.io/usage for more information.
        """
    )

_spacy_version = get_pkg_version("spacy")

_check_compat = _spacy_version.startswith("3")
if not _check_compat:  # pragma: no cover
    raise EnvironmentError(
        "BentoML will only provide supports for spacy 3.x and above"
        " as we can provide more supports for Spacy new design. Currently"
        f" detected spacy to have version {_spacy_version}"
    )

util = LazyLoader("util", globals(), "spacy.util")
thinc_util = LazyLoader("thinc_util", globals(), "thinc.util")
thinc_backends = LazyLoader("thinc_backends", globals(), "thinc.backends")

torch = LazyLoader("torch", globals(), "torch")
tensorflow = LazyLoader("tensorflow", globals(), "tensorflow")

_TORCH_TF_WARNING = """\
It is recommended that if you want to run SpaCy with {framework}
 model that you also have `{package}` installed.
Refers to {link} for more information.

We also detected that you choose to run on GPUs, thus in order to utilize
 BentoML Runners features with GPUs you should also install `{package}` with
 CUDA support.
"""  # noqa

PROJECTS_CMD_NOT_SUPPORTED = [
    "assets",
    "document",
    "dvc",
    "push",
    "run",
]

logger = logging.getLogger(__name__)


@inject
def load_project(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> str:
    model = model_store.get(tag)
    if "projects_uri" in model.info.options:
        logger.warning(
            "We will only returns the path of projects saved under BentoML modelstore"
            " and leave projects interaction for users to decide what they want to do."
            " Refers to https://spacy.io/api/cli#project for more information."
        )
        return os.path.join(model.path, model.info.options["target_path"])
    raise EnvironmentError(
        "Cannot use `bentoml.spacy.load_project()` to load non Spacy Projects. If your"
        " model is not a Spacy projects use `bentoml.spacy.load()` instead."
    )


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    vocab: t.Union["Vocab", bool] = True,  # type: ignore[reportUnknownParameterType]
    disable: t.Iterable[str] = util.SimpleFrozenList(),  # noqa
    exclude: t.Iterable[str] = util.SimpleFrozenList(),  # noqa
    config: t.Union[t.Dict[str, t.Any], "Config"] = util.SimpleFrozenDict(),  # noqa
) -> "spacy.language.Language":
    model = model_store.get(tag)
    if "projects_uri" in model.info.options:
        raise EnvironmentError(
            "Cannot use `bentoml.spacy.load()` to load Spacy Projects. Use"
            " `bentoml.spacy.load_project()` instead."
        )
    required = model.info.options["pip_package"]

    try:
        _ = importlib.import_module(required)
    except ModuleNotFoundError:
        try:
            from spacy.cli.download import download

            # TODO move this to runner on startup hook
            download(required)
        except (SystemExit, Exception):  # pylint: disable=broad-except
            logger.warning(
                f"{required} cannot be downloaded as pip package. If this"
                f" is a custom pipeline there is nothing to worry about."
                f" If this is a pretrained model provided by Explosion make"
                f" sure that you save the correct package and model to BentoML"
                f" via `bentoml.spacy.save()`"
            )
    try:
        # check if pipeline has additional requirements then all related
        # pip package has been installed correctly.
        additional = model.info.options["additional_requirements"]
        not_existed = list()  # type: t.List[str]
        dists = packages_distributions()
        for module_name in additional:
            mod, _ = split_requirement(module_name)
            if mod not in dists:
                not_existed.append(module_name)
            if len(not_existed) > 0:
                raise MissingDependencyException(
                    f"`{','.join(not_existed)}` is required by `{tag}`."
                )
    except KeyError:
        pass
    return util.load_model(
        model.path, vocab=vocab, disable=disable, exclude=exclude, config=config
    )


@inject
def save(
    name: str,
    model: "spacy.language.Language",
    *,
    metadata: t.Union[None, t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (`str`):
            Name for given model instance. This should pass Python identifier check.
        model (`spacy.language.Language`):
            Instance of model to be saved
        metadata (`t.Optional[t.Dict[str, t.Any]]`, default to `None`):
            Custom metadata for given model.
        model_store (`~bentoml._internal.models.store.ModelStore`, default to `BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        tag (`str` with a format `name:version`) where `name` is the defined name user
        set for their models, and version will be generated by BentoML.

    Examples::
        import spacy
        import bentoml.spacy
        nlp = spacy.load("en_core_web_trf")

        # custom training or layers here

        bentoml.spacy.save("spacy_roberta", nlp)
    """  # noqa
    context: t.Dict[str, t.Any] = {
        "framework_name": "spacy",
        "pip_dependencies": [f"spacy=={_spacy_version}"],
    }
    _model = Model.create(
        name,
        module=__name__,
        options=None,
        context=context,
        metadata=metadata,
    )

    meta = model.meta
    pip_package = f"{meta['lang']}_{meta['name']}"
    _model.info.options = {"pip_package": pip_package}
    if "requirements" in meta:
        _model.info.options["additional_requirements"] = meta["requirements"]
    model.to_disk(_model.path)

    _model.save(model_store)
    return _model.tag


@inject
def projects(
    save_name: str,
    tasks: str,
    name: t.Optional[str] = None,
    repo_or_store: t.Optional[str] = None,
    remotes_config: t.Optional[t.Dict[str, t.Dict[str, str]]] = None,
    *,
    branch: t.Optional[str] = None,
    sparse_checkout: bool = False,
    verbose: bool = True,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    from spacy.cli.project.pull import project_pull
    from spacy.cli.project.clone import project_clone

    if repo_or_store is None:
        repo_or_store = getattr(spacy.about, "__projects__")
    if branch is None:
        branch = getattr(spacy.about, "__projects_branch__")

    if tasks in PROJECTS_CMD_NOT_SUPPORTED:
        raise BentoMLException(
            """\
        BentoML only supports `clone` and `pull` for
         git and remote storage from SpaCy CLI to
         save into modelstore. Refers to
         https://spacy.io/api/cli#project for more
         information on SpaCy Projects.
        """
        )
    context: t.Dict[str, t.Any] = {
        "framework_name": "spacy",
        "pip_dependencies": [f"spacy=={_spacy_version}"],
        "tasks": tasks,
    }
    _model = Model.create(
        save_name,
        module=__name__,
        options=None,
        context=context,
        metadata=metadata,
    )
    output_path = _model.path_of(SAVE_NAMESPACE)
    _model.info.options = {"projects_uri": repo_or_store, "target_path": SAVE_NAMESPACE}
    if tasks == "clone":
        # TODO: update check for master or main branch
        assert (
            name is not None
        ), "`name` of the template is required to clone a project."
        _model.info.options["name"] = name
        assert isinstance(repo_or_store, str) and isinstance(branch, str)
        project_clone(
            name,
            Path(output_path),
            repo=repo_or_store,
            branch=branch,
            sparse_checkout=sparse_checkout,
        )
        copy_tree(_model.path, output_path)
    else:
        # works with S3 bucket, haven't failed yet
        assert (
            remotes_config is not None
        ), """\
            `remotes_config` is required in order to pull projects into
             BentoML modelstore. Refers to
             https://spacy.io/usage/projects#remote
             for more information. We will accept remotes
             as shown:
             {
                'remotes':
                {
                    'default':'s3://spacy-bucket',
                }
            }
             """
        os.makedirs(output_path, exist_ok=True)
        with Path(output_path, "project.yml").open("w") as inf:
            yaml.dump(remotes_config, inf)
        for remote in remotes_config.get("remotes", {}):
            pull: "partial[t.Generator[t.Tuple[str, str], None, None]]" = partial(
                project_pull, remote=remote, verbose=verbose
            )
            for url, res_path in pull(Path(output_path)):
                if url is not None:  # pragma: no cover
                    logger.info(f"Pulled {res_path} from {repo_or_store}")
    _model.save(model_store)
    return _model.tag


class _SpacyRunner(Runner):
    @inject
    def __init__(
        self,
        tag: t.Union[str, Tag],
        gpu_device_id: t.Optional[int],
        resource_quota: t.Optional[t.Dict[str, t.Any]],
        batch_options: t.Optional[t.Dict[str, t.Any]],
        vocab: t.Union["Vocab", bool],  # type: ignore[reportUnknownParameterType]
        disable: t.Iterable[str],
        exclude: t.Iterable[str],
        config: t.Union[t.Dict[str, t.Any], "Config"],
        as_tuples: t.Union[Literal[True], Literal[False]],
        batch_size: t.Optional[int],
        component_cfg: t.Optional[t.Dict[str, t.Dict[str, t.Any]]],
        backend_options: t.Optional[Literal["pytorch", "tensorflow"]] = "pytorch",
        model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    ):
        in_store_tag = model_store.get(tag).tag
        super().__init__(str(in_store_tag), resource_quota, batch_options)
        self._tag = in_store_tag

        self._vocab: t.Union["Vocab", bool] = vocab
        self._disable = disable
        self._exclude = exclude
        self._config = config

        self._model_store = model_store
        self._backend_options = backend_options
        self._gpu_device_id = gpu_device_id

        if self._gpu_device_id is not None:
            if resource_quota is None:
                resource_quota = dict(gpus=self._gpu_device_id)
            else:
                resource_quota["gpus"] = self._gpu_device_id
        self._configure(backend_options)
        self._as_tuples = as_tuples
        self._batch_size = batch_size
        self._component_cfg = component_cfg
        super().__init__(str(tag), resource_quota, batch_options)

    def _configure(self, backend_options: t.Optional[str]) -> None:
        if self._gpu_device_id is not None and thinc_util.prefer_gpu(
            self._gpu_device_id
        ):  # pragma: no cover
            assert backend_options is not None
            if backend_options == "pytorch":
                thinc_backends.use_pytorch_for_gpu_memory()
            else:
                thinc_backends.use_tensorflow_for_gpu_memory()
            thinc_util.require_gpu(self._gpu_device_id)
            thinc_backends.set_gpu_allocator(backend_options)
            thinc_util.set_active_gpu(self._gpu_device_id)
        else:
            thinc_util.require_cpu()

    @property
    def required_models(self) -> t.List[Tag]:
        return [self._tag]

    @property
    def num_concurrency_per_replica(self) -> int:
        if self.resource_quota.on_gpu:
            return 1
        return int(round(self.resource_quota.cpu))

    def _get_pytorch_gpu_count(self) -> t.Optional[int]:
        assert self._backend_options == "pytorch"
        devs = getattr(torch, "cuda").device_count()
        if devs == 0:
            logger.warning("Installation of Torch is not CUDA-enabled.")
            logger.warning(
                _TORCH_TF_WARNING.format(
                    framework="PyTorch",
                    package="torch",
                    link="https://pytorch.org/get-started/locally/",
                )
            )
            return None
        return devs

    def _get_tensorflow_gpu_count(self) -> t.Optional[int]:
        assert self._backend_options == "tensorflow"
        from tensorflow.python.client import (
            device_lib,  # type: ignore; pylint: disable=E0611
        )

        try:
            return len(
                [
                    x
                    for x in getattr(device_lib, "list_local_devices")()
                    if getattr(x, "device_type") == "GPU"
                ]
            )
        except (AttributeError, Exception):  # pylint: disable=broad-except
            logger.warning(
                _TORCH_TF_WARNING.format(
                    framework="Tensorflow 2.x",
                    package="tensorflow",
                    link="https://www.tensorflow.org/install/gpu",
                )
            )
            return None

    @property
    def num_replica(self) -> int:
        if self.resource_quota.on_gpu:
            if self._backend_options == "pytorch":
                num_devices = self._get_pytorch_gpu_count()
            else:
                num_devices = self._get_tensorflow_gpu_count()
            return num_devices if num_devices is not None else 1
        return 1

    # pylint: disable=arguments-differ,attribute-defined-outside-init
    def _setup(self) -> None:
        self._model = load(
            self._tag,
            model_store=self._model_store,
            vocab=self._vocab,
            exclude=self._exclude,
            disable=self._disable,
            config=self._config,
        )

    # pylint: disable=arguments-differ
    def _run_batch(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        input_data: t.Union[t.Iterable[t.Tuple[t.Union[str, "Doc"], t.Any]], t.Union[str, "Doc"]],  # type: ignore[reportUnknownParameterType] # noqa: LN001
    ) -> t.Union[t.Iterator["Doc"], t.Iterator[t.Tuple["Doc", t.Any]]]:  # type: ignore[reportUnknownParameterType] # noqa: LN001
        return self._model.pipe(  # type: ignore[reportGeneralTypeIssues]
            input_data,  # type: ignore[reportGeneralTypeIssues]
            as_tuples=self._as_tuples,
            batch_size=self._batch_size,
            disable=self._disable,
            component_cfg=self._component_cfg,
            n_process=self.num_replica,
        )


@inject
def load_runner(
    tag: t.Union[str, Tag],
    *,
    gpu_device_id: t.Optional[int] = None,
    backend_options: t.Optional[Literal["pytorch", "tensorflow"]] = None,
    resource_quota: t.Optional[t.Dict[str, t.Any]] = None,
    batch_options: t.Optional[t.Dict[str, t.Any]] = None,
    vocab: t.Union["Vocab", bool] = True,  # type: ignore[reportUnknownParameterType]
    disable: t.Iterable[str] = util.SimpleFrozenList(),  # noqa
    exclude: t.Iterable[str] = util.SimpleFrozenList(),  # noqa
    config: t.Union[t.Dict[str, t.Any], "Config"] = util.SimpleFrozenDict(),  # noqa
    as_tuples: t.Union[Literal[True], Literal[False]] = False,
    batch_size: t.Optional[int] = None,
    component_cfg: t.Optional[t.Dict[str, t.Dict[str, t.Any]]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "_SpacyRunner":
    return _SpacyRunner(
        tag=tag,
        gpu_device_id=gpu_device_id,
        backend_options=backend_options,
        resource_quota=resource_quota,
        batch_options=batch_options,
        vocab=vocab,
        disable=disable,
        exclude=exclude,
        config=config,
        as_tuples=as_tuples,
        batch_size=batch_size,
        component_cfg=component_cfg,
        model_store=model_store,
    )
